#!/usr/bin/env python
# License: GPLv3 Copyright: 2015, Kovid Goyal <kovid at kovidgoyal.net>

import os
import stat
from functools import partial
from threading import Event, Thread

from qt.core import (
    QAbstractItemView,
    QBoxLayout,
    QDialog,
    QDialogButtonBox,
    QFrame,
    QGridLayout,
    QIcon,
    QLabel,
    QLineEdit,
    QListWidget,
    QListWidgetItem,
    QProgressBar,
    QPushButton,
    QScrollArea,
    QSize,
    QStackedLayout,
    Qt,
    QVBoxLayout,
    QWidget,
    pyqtSignal,
)

from calibre import as_unicode, human_readable
from calibre.constants import iswindows
from calibre.db.legacy import LibraryDatabase
from calibre.gui2 import choose_dir, error_dialog, question_dialog
from calibre.gui2.widgets2 import Dialog
from calibre.startup import connect_lambda
from calibre.utils.exim import Importer, all_known_libraries, export, import_data
from calibre.utils.icu import numeric_sort_key
from calibre.utils.localization import _


def disk_usage(path_to_dir, abort=None):
    stack = [path_to_dir]
    ans = 0
    while stack:
        bdir = stack.pop()
        try:
            for child in os.listdir(bdir):
                cpath = os.path.join(bdir, child)
                if abort is not None and abort.is_set():
                    return -1
                r = os.lstat(cpath)
                if stat.S_ISDIR(r.st_mode):
                    stack.append(cpath)
                ans += r.st_size
        except OSError:
            pass
    return ans


class ImportLocation(QWidget):
    def __init__(self, lpath, parent=None):
        QWidget.__init__(self, parent)
        self.l = l = QGridLayout(self)
        self.la = la = QLabel(_('Previous location: ') + lpath)
        la.setWordWrap(True)
        self.lpath = lpath
        l.addWidget(la, 0, 0, 1, -1)
        self.le = le = QLineEdit(self)
        le.setPlaceholderText(_('Location to import this library to'))
        l.addWidget(le, 1, 0)
        self.b = b = QPushButton(QIcon.ic('document_open.png'), _('Select &folder'), self)
        b.clicked.connect(self.select_folder)
        l.addWidget(b, 1, 1)
        self.lpath = lpath

    def select_folder(self):
        path = choose_dir(self, 'select-folder-for-imported-library', _('Choose a folder for this library'))
        if path is not None:
            self.le.setText(path)

    @property
    def path(self):
        return self.le.text().strip()


class RunAction(QDialog):
    update_current_signal = pyqtSignal(object, object, object)
    update_overall_signal = pyqtSignal(object, object, object)
    finish_signal = pyqtSignal()

    def __init__(self, title, err_msg, action, parent=None):
        QDialog.__init__(self, parent)
        self.setWindowTitle(_('Working, please wait...'))
        self.title, self.action, self.tb, self.err_msg = title, action, None, err_msg
        self.abort = Event()
        self.setup_ui()
        t = Thread(name='ExImWorker', target=self.run_action)
        t.daemon = True
        t.start()

    def setup_ui(self):
        self.l = l = QGridLayout(self)
        self.bb = QDialogButtonBox(self)
        self.bb.setStandardButtons(QDialogButtonBox.StandardButton.Cancel)
        self.bb.rejected.connect(self.reject)

        self.la1 = la = QLabel('<h2>' + self.title)
        l.addWidget(la, 0, 0, 1, -1)
        self.la2 = la = QLabel(_('Total:'))
        l.addWidget(la, l.rowCount(), 0)
        self.overall = p = QProgressBar(self)
        p.setMinimum(0), p.setValue(0), p.setMaximum(0)
        p.setMinimumWidth(450)
        l.addWidget(p, l.rowCount() - 1, 1)
        self.omsg = la = QLabel(self)
        la.setMaximumWidth(450)
        l.addWidget(la, l.rowCount(), 1)
        self.la3 = la = QLabel(_('Current:'))
        l.addWidget(la, l.rowCount(), 0)
        self.current = p = QProgressBar(self)
        p.setMinimum(0), p.setValue(0), p.setMaximum(0)
        l.addWidget(p, l.rowCount() - 1, 1)
        self.cmsg = la = QLabel(self)
        la.setMaximumWidth(450)
        l.addWidget(la, l.rowCount(), 1)
        l.addWidget(self.bb, l.rowCount(), 0, 1, -1)
        self.update_current_signal.connect(self.update_current, type=Qt.ConnectionType.QueuedConnection)
        self.update_overall_signal.connect(self.update_overall, type=Qt.ConnectionType.QueuedConnection)
        self.finish_signal.connect(self.finish_processing, type=Qt.ConnectionType.QueuedConnection)

    def update_overall(self, msg, count, total):
        self.overall.setMaximum(total), self.overall.setValue(count)
        self.omsg.setText(msg)

    def update_current(self, msg, count, total):
        self.current.setMaximum(total), self.current.setValue(count)
        self.cmsg.setText(msg)

    def reject(self):
        self.abort.set()
        cancel_btn = self.bb.button(QDialogButtonBox.StandardButton.Cancel)
        assert cancel_btn is not None
        cancel_btn.setEnabled(False)

    def finish_processing(self):
        if self.abort.is_set():
            return QDialog.reject(self)
        if self.tb is not None:
            error_dialog(
                self,
                _('Failed'),
                self.err_msg + ' ' + _('Click "Show details" for more information.'),
                det_msg=self.tb,
                show=True,
            )
        self.accept()

    def run_action(self):
        try:
            self.action(abort=self.abort, progress1=self.update_overall_signal.emit, progress2=self.update_current_signal.emit)
        except Exception:
            import traceback

            self.tb = traceback.format_exc()
        self.finish_signal.emit()


class EximDialog(Dialog):
    update_disk_usage = pyqtSignal(object, object)

    def __init__(self, parent=None, initial_panel=None):
        self.initial_panel = initial_panel
        self.abort_disk_usage = Event()
        self.restart_needed = False
        Dialog.__init__(self, _('Export/import all calibre data'), 'exim-calibre', parent=parent)

    def sizeHint(self):
        return QSize(800, 600)

    def setup_ui(self):
        self.l = l = QVBoxLayout(self)
        self.stack = s = QStackedLayout()
        l.addLayout(s)
        l.addWidget(self.bb)
        self.welcome = w = QWidget(self)
        s.addWidget(w)
        l = QVBoxLayout(w)
        la = QLabel(
            '<p>'
            + _(
                'You can export all calibre data, including your books, settings and plugins'
                ' into a single folder. Then, you can use this tool to re-import all that'
                ' data into a different calibre install, for example, on another computer.'
            )
            + '<p>'
            + _(
                'This is a simple way to move your calibre installation with all its data to'
                ' a new computer, or to replicate your current setup on a second computer.'
            )
        )
        la.setWordWrap(True)
        l.addWidget(la)
        l.addSpacing(20)
        self.exp_button = b = QPushButton(_('&Export all your calibre data'))
        connect_lambda(b.clicked, self, lambda self: self.show_panel('export'))
        l.addWidget(b), l.addSpacing(20)
        self.imp_button = b = QPushButton(_('&Import previously exported data'))
        connect_lambda(b.clicked, self, lambda self: self.show_panel('import'))
        l.addWidget(b), l.addStretch(20)

        self.setup_export_panel()
        self.setup_import_panel()
        self.show_panel(self.initial_panel)

    def export_lib_text(self, lpath, size=None):
        return _('{0} [Size: {1}]\nin {2}').format(
            os.path.basename(lpath),
            ('' if size < 0 else human_readable(size)) if size is not None else _('Calculating...'),
            os.path.dirname(lpath),
        )

    def setup_export_panel(self):
        self.export_panel = w = QWidget(self)
        self.stack.addWidget(w)
        l = QVBoxLayout(w)
        la = QLabel(_('Select which libraries you want to export below'))
        la.setWordWrap(True), l.addWidget(la)
        self.lib_list = ll = QListWidget(self)
        l.addWidget(ll)
        ll.setSelectionMode(QAbstractItemView.SelectionMode.ExtendedSelection)
        ll.setStyleSheet('QListView::item { padding: 5px }')
        ll.setAlternatingRowColors(True)
        lpaths = all_known_libraries()
        for lpath in sorted(lpaths, key=lambda x: numeric_sort_key(os.path.basename(x))):
            i = QListWidgetItem(self.export_lib_text(lpath), ll)
            i.setData(Qt.ItemDataRole.UserRole, lpath)
            i.setData(Qt.ItemDataRole.UserRole + 1, lpaths[lpath])
            i.setIcon(QIcon.ic('lt.png'))
            i.setSelected(True)

        def _on_disk_usage_update(i, sz):
            item = self.lib_list.item(i)
            assert item is not None
            item.setText(self.export_lib_text(item.data(Qt.ItemDataRole.UserRole), sz))

        self.update_disk_usage.connect(_on_disk_usage_update, type=Qt.ConnectionType.QueuedConnection)

    def get_lib_sizes(self):
        for i in range(self.lib_list.count()):
            lib_list_item = self.lib_list.item(i)
            assert lib_list_item is not None
            path = lib_list_item.data(Qt.ItemDataRole.UserRole)
            try:
                sz = disk_usage(path, abort=self.abort_disk_usage)
            except Exception:
                import traceback

                traceback.print_exc()
            self.update_disk_usage.emit(i, sz)

    def setup_import_panel(self):
        self.import_panel = w = QWidget(self)
        self.stack.addWidget(w)
        self.import_panel_stack = s = QStackedLayout(w)
        self.ig = w = QWidget()
        s.addWidget(w)
        l = QVBoxLayout(w)
        la = QLabel(_('<b style="color: red">WARNING</b>: do not import calibre data you get from an untrusted source, as that is a security risk.'))
        la.setWordWrap(True)
        l.addWidget(la)
        la = QLabel(_('Specify the folder containing the previously exported calibre data that you wish to import.'))
        la.setWordWrap(True)
        l.addWidget(la)
        self.export_dir_button = b = QPushButton(QIcon.ic('document_open.png'), _('Choose &folder'), self)
        b.clicked.connect(self.select_import_folder)
        l.addWidget(b), l.addStretch()

        self.select_libraries_panel = w = QScrollArea(self)
        w.setWidgetResizable(True)
        s.addWidget(w)
        self.slp = w = QWidget(self)
        self.select_libraries_panel.setWidget(w)
        l = QVBoxLayout(w)
        la = QLabel(
            _(
                'Specify locations for the libraries you want to import. A location must be an empty folder'
                ' on your computer. If you leave any blank, those libraries will not be imported.'
            )
        )
        la.setWordWrap(True)
        l.addWidget(la)

    def select_import_folder(self):
        path = choose_dir(self, 'choose-export-folder-for-import', _('Select folder with exported data'))
        if path is None:
            return
        if not question_dialog(
            self,
            _('Are you sure?'),
            _(
                'Importing calibre data means all libraries, settings, plugins, etc will be imported. This is'
                ' a security risk, only proceed if the data you are importing was previously generated by you, using the calibre'
                ' export functionality.'
            ),
        ):
            return
        try:
            self.importer = Importer(path)
        except Exception as e:
            import traceback

            return error_dialog(
                self,
                _('Not valid'),
                _('The folder {0} is not valid: {1}').format(path, as_unicode(e)),
                det_msg=traceback.format_exc(),
                show=True,
            )
        self.setup_select_libraries_panel()
        self.import_panel_stack.setCurrentIndex(1)

    def setup_select_libraries_panel(self):
        self.imported_lib_widgets = []
        self.frames = []
        l = self.slp.layout()
        assert isinstance(l, QBoxLayout)
        for lpath in sorted(self.importer.metadata['libraries'], key=lambda x: numeric_sort_key(os.path.basename(x))):
            f = QFrame(self)
            self.frames.append(f)
            l.addWidget(f)
            f.setFrameShape(QFrame.Shape.HLine)
            w = ImportLocation(lpath, self.slp)
            l.addWidget(w)
            self.imported_lib_widgets.append(w)
        l.addStretch()

    def validate_import(self):
        from calibre.gui2.ui import get_gui

        g = get_gui()
        if g is not None:
            if g.iactions['Connect Share'].content_server_is_running:
                error_dialog(
                    self,
                    _('Content server running'),
                    _('Cannot import while the Content server is running, shut it down first by clicking the "Connect/share" button on the calibre toolbar'),
                    show=True,
                )
                return False
        if self.import_panel_stack.currentIndex() == 0:
            error_dialog(
                self,
                _('No folder selected'),
                _('You must select a folder containing the previously exported data that you wish to import'),
                show=True,
            )
            return False
        else:
            blanks = []
            for w in self.imported_lib_widgets:
                newloc = w.path
                if not newloc:
                    blanks.append(w.lpath)
                    continue
                if iswindows and len(newloc) > LibraryDatabase.WINDOWS_LIBRARY_PATH_LIMIT:
                    error_dialog(
                        self,
                        _('Too long'),
                        _('Path to library ({0}) too long. It must be less than {1} characters.').format(newloc, LibraryDatabase.WINDOWS_LIBRARY_PATH_LIMIT),
                        show=True,
                    )
                    return False
                if not os.path.isdir(newloc):
                    error_dialog(self, _('Not a folder'), _('%s is not a folder') % newloc, show=True)
                    return False
                if os.listdir(newloc):
                    error_dialog(self, _('Folder not empty'), _('%s is not an empty folder') % newloc, show=True)
                    return False
            if blanks:
                if len(blanks) == len(self.imported_lib_widgets):
                    error_dialog(
                        self,
                        _('No libraries selected'),
                        _('You must specify the location for at least one library'),
                        show=True,
                    )
                    return False
                if not question_dialog(
                    self,
                    _('Some libraries ignored'),
                    _('You have chosen not to import some libraries. Proceed anyway?'),
                ):
                    return False
        return True

    def show_panel(self, which):
        self.validate = self.run_action = lambda: True
        if which is None:
            self.bb.setStandardButtons(QDialogButtonBox.StandardButton.Cancel)
        else:
            if which == 'export':
                self.validate = self.validate_export
                self.run_action = self.run_export_action
                t = Thread(name='GetLibSizes', target=self.get_lib_sizes)
                t.daemon = True
                t.start()
            else:
                self.validate = self.validate_import
                self.run_action = self.run_import_action
            self.bb.setStandardButtons(QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel)
        self.stack.setCurrentIndex({'export': 1, 'import': 2}.get(which, 0))

    def validate_export(self):
        path = choose_dir(self, 'export-calibre-dir', _('Choose a folder to export to'))
        if not path:
            return False
        if os.listdir(path):
            error_dialog(
                self,
                _('Export folder not empty'),
                _('The folder you choose to export the data to must be empty.'),
                show=True,
            )
            return False
        self.export_dir = path
        return True

    def run_export_action(self):
        from calibre.gui2.ui import get_gui

        library_paths = {i.data(Qt.ItemDataRole.UserRole): i.data(Qt.ItemDataRole.UserRole + 1) for i in self.lib_list.selectedItems()}
        dbmap = {}
        gui = get_gui()
        if gui is not None:
            db = gui.current_db
            dbmap[db.library_path] = db.new_api
        return (
            RunAction(
                _('Exporting all calibre data...'),
                _('Failed to export data.'),
                partial(export, self.export_dir, library_paths=library_paths, dbmap=dbmap),
                parent=self,
            ).exec()
            == QDialog.DialogCode.Accepted
        )

    def run_import_action(self):
        library_path_map = {}
        for w in self.imported_lib_widgets:
            if w.path:
                library_path_map[w.lpath] = w.path
        return (
            RunAction(
                _('Importing all calibre data...'),
                _('Failed to import data.'),
                partial(import_data, self.importer, library_path_map),
                parent=self,
            ).exec()
            == QDialog.DialogCode.Accepted
        )

    def accept(self):
        if not self.validate():
            return
        self.abort_disk_usage.set()
        if self.run_action():
            self.restart_needed = self.stack.currentIndex() == 2
            Dialog.accept(self)

    def reject(self):
        self.abort_disk_usage.set()
        Dialog.reject(self)


if __name__ == '__main__':
    from calibre.gui2 import Application

    app = Application([])
    d = EximDialog(initial_panel='import')
    d.exec()
    del app



#!/usr/bin/env python
# Copyright (C) 2008 Søren Roug, European Environment Agency
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Contributor(s):
#


from .namespaces import (
    ANIMNS,
    CHARTNS,
    DR3DNS,
    DRAWNS,
    FORMNS,
    MANIFESTNS,
    METANS,
    NUMBERNS,
    OFFICENS,
    PRESENTATIONNS,
    SCRIPTNS,
    STYLENS,
    SVGNS,
    TABLENS,
    TEXTNS,
)

# Inline element don't cause a box
# They are analogous to the HTML elements SPAN, B, I etc.
inline_elements = (
    (TEXTNS, 'a'),
    (TEXTNS, 'author-initials'),
    (TEXTNS, 'author-name'),
    (TEXTNS, 'bibliography-mark'),
    (TEXTNS, 'bookmark-ref'),
    (TEXTNS, 'chapter'),
    (TEXTNS, 'character-count'),
    (TEXTNS, 'conditional-text'),
    (TEXTNS, 'creation-date'),
    (TEXTNS, 'creation-time'),
    (TEXTNS, 'creator'),
    (TEXTNS, 'database-display'),
    (TEXTNS, 'database-name'),
    (TEXTNS, 'database-next'),
    (TEXTNS, 'database-row-number'),
    (TEXTNS, 'database-row-select'),
    (TEXTNS, 'date'),
    (TEXTNS, 'dde-connection'),
    (TEXTNS, 'description'),
    (TEXTNS, 'editing-cycles'),
    (TEXTNS, 'editing-duration'),
    (TEXTNS, 'execute-macro'),
    (TEXTNS, 'expression'),
    (TEXTNS, 'file-name'),
    (TEXTNS, 'hidden-paragraph'),
    (TEXTNS, 'hidden-text'),
    (TEXTNS, 'image-count'),
    (TEXTNS, 'initial-creator'),
    (TEXTNS, 'keywords'),
    (TEXTNS, 'measure'),
    (TEXTNS, 'modification-date'),
    (TEXTNS, 'modification-time'),
    (TEXTNS, 'note-ref'),
    (TEXTNS, 'object-count'),
    (TEXTNS, 'page-continuation'),
    (TEXTNS, 'page-count'),
    (TEXTNS, 'page-number'),
    (TEXTNS, 'page-variable-get'),
    (TEXTNS, 'page-variable-set'),
    (TEXTNS, 'paragraph-count'),
    (TEXTNS, 'placeholder'),
    (TEXTNS, 'print-date'),
    (TEXTNS, 'printed-by'),
    (TEXTNS, 'print-time'),
    (TEXTNS, 'reference-ref'),
    (TEXTNS, 'ruby'),
    (TEXTNS, 'ruby-base'),
    (TEXTNS, 'ruby-text'),
    (TEXTNS, 'script'),
    (TEXTNS, 'sender-city'),
    (TEXTNS, 'sender-company'),
    (TEXTNS, 'sender-country'),
    (TEXTNS, 'sender-email'),
    (TEXTNS, 'sender-fax'),
    (TEXTNS, 'sender-firstname'),
    (TEXTNS, 'sender-initials'),
    (TEXTNS, 'sender-lastname'),
    (TEXTNS, 'sender-phone-private'),
    (TEXTNS, 'sender-phone-work'),
    (TEXTNS, 'sender-position'),
    (TEXTNS, 'sender-postal-code'),
    (TEXTNS, 'sender-state-or-province'),
    (TEXTNS, 'sender-street'),
    (TEXTNS, 'sender-title'),
    (TEXTNS, 'sequence'),
    (TEXTNS, 'sequence-ref'),
    (TEXTNS, 'sheet-name'),
    (TEXTNS, 'span'),
    (TEXTNS, 'subject'),
    (TEXTNS, 'table-count'),
    (TEXTNS, 'table-formula'),
    (TEXTNS, 'template-name'),
    (TEXTNS, 'text-input'),
    (TEXTNS, 'time'),
    (TEXTNS, 'title'),
    (TEXTNS, 'user-defined'),
    (TEXTNS, 'user-field-get'),
    (TEXTNS, 'user-field-input'),
    (TEXTNS, 'variable-get'),
    (TEXTNS, 'variable-input'),
    (TEXTNS, 'variable-set'),
    (TEXTNS, 'word-count'),
)


# It is almost impossible to determine what elements are block elements.
# There are so many that don't fit the form
block_elements = (
    (TEXTNS, 'h'),
    (TEXTNS, 'p'),
    (TEXTNS, 'list'),
    (TEXTNS, 'list-item'),
    (TEXTNS, 'section'),
)

declarative_elements = (
    (OFFICENS, 'font-face-decls'),
    (PRESENTATIONNS, 'date-time-decl'),
    (PRESENTATIONNS, 'footer-decl'),
    (PRESENTATIONNS, 'header-decl'),
    (TABLENS, 'table-template'),
    (TEXTNS, 'alphabetical-index-entry-template'),
    (TEXTNS, 'alphabetical-index-source'),
    (TEXTNS, 'bibliography-entry-template'),
    (TEXTNS, 'bibliography-source'),
    (TEXTNS, 'dde-connection-decls'),
    (TEXTNS, 'illustration-index-entry-template'),
    (TEXTNS, 'illustration-index-source'),
    (TEXTNS, 'index-source-styles'),
    (TEXTNS, 'index-title-template'),
    (TEXTNS, 'note-continuation-notice-backward'),
    (TEXTNS, 'note-continuation-notice-forward'),
    (TEXTNS, 'notes-configuration'),
    (TEXTNS, 'object-index-entry-template'),
    (TEXTNS, 'object-index-source'),
    (TEXTNS, 'sequence-decls'),
    (TEXTNS, 'table-index-entry-template'),
    (TEXTNS, 'table-index-source'),
    (TEXTNS, 'table-of-content-entry-template'),
    (TEXTNS, 'table-of-content-source'),
    (TEXTNS, 'user-field-decls'),
    (TEXTNS, 'user-index-entry-template'),
    (TEXTNS, 'user-index-source'),
    (TEXTNS, 'variable-decls'),
)

empty_elements = (
    (ANIMNS, 'animate'),
    (ANIMNS, 'animateColor'),
    (ANIMNS, 'animateMotion'),
    (ANIMNS, 'animateTransform'),
    (ANIMNS, 'audio'),
    (ANIMNS, 'param'),
    (ANIMNS, 'set'),
    (ANIMNS, 'transitionFilter'),
    (CHARTNS, 'categories'),
    (CHARTNS, 'data-point'),
    (CHARTNS, 'domain'),
    (CHARTNS, 'error-indicator'),
    (CHARTNS, 'floor'),
    (CHARTNS, 'grid'),
    (CHARTNS, 'legend'),
    (CHARTNS, 'mean-value'),
    (CHARTNS, 'regression-curve'),
    (CHARTNS, 'stock-gain-marker'),
    (CHARTNS, 'stock-loss-marker'),
    (CHARTNS, 'stock-range-line'),
    (CHARTNS, 'symbol-image'),
    (CHARTNS, 'wall'),
    (DR3DNS, 'cube'),
    (DR3DNS, 'extrude'),
    (DR3DNS, 'light'),
    (DR3DNS, 'rotate'),
    (DR3DNS, 'sphere'),
    (DRAWNS, 'contour-path'),
    (DRAWNS, 'contour-polygon'),
    (DRAWNS, 'equation'),
    (DRAWNS, 'fill-image'),
    (DRAWNS, 'floating-frame'),
    (DRAWNS, 'glue-point'),
    (DRAWNS, 'gradient'),
    (DRAWNS, 'handle'),
    (DRAWNS, 'hatch'),
    (DRAWNS, 'layer'),
    (DRAWNS, 'marker'),
    (DRAWNS, 'opacity'),
    (DRAWNS, 'page-thumbnail'),
    (DRAWNS, 'param'),
    (DRAWNS, 'stroke-dash'),
    (FORMNS, 'connection-resource'),
    (FORMNS, 'list-value'),
    (FORMNS, 'property'),
    (MANIFESTNS, 'algorithm'),
    (MANIFESTNS, 'key-derivation'),
    (METANS, 'auto-reload'),
    (METANS, 'document-statistic'),
    (METANS, 'hyperlink-behaviour'),
    (METANS, 'template'),
    (NUMBERNS, 'am-pm'),
    (NUMBERNS, 'boolean'),
    (NUMBERNS, 'day'),
    (NUMBERNS, 'day-of-week'),
    (NUMBERNS, 'era'),
    (NUMBERNS, 'fraction'),
    (NUMBERNS, 'hours'),
    (NUMBERNS, 'minutes'),
    (NUMBERNS, 'month'),
    (NUMBERNS, 'quarter'),
    (NUMBERNS, 'scientific-number'),
    (NUMBERNS, 'seconds'),
    (NUMBERNS, 'text-content'),
    (NUMBERNS, 'week-of-year'),
    (NUMBERNS, 'year'),
    (OFFICENS, 'dde-source'),
    (PRESENTATIONNS, 'date-time'),
    (PRESENTATIONNS, 'footer'),
    (PRESENTATIONNS, 'header'),
    (PRESENTATIONNS, 'placeholder'),
    (PRESENTATIONNS, 'play'),
    (PRESENTATIONNS, 'show'),
    (PRESENTATIONNS, 'sound'),
    (SCRIPTNS, 'event-listener'),
    (STYLENS, 'column'),
    (STYLENS, 'column-sep'),
    (STYLENS, 'drop-cap'),
    (STYLENS, 'footnote-sep'),
    (STYLENS, 'list-level-properties'),
    (STYLENS, 'map'),
    (STYLENS, 'ruby-properties'),
    (STYLENS, 'table-column-properties'),
    (STYLENS, 'tab-stop'),
    (STYLENS, 'text-properties'),
    (SVGNS, 'definition-src'),
    (SVGNS, 'font-face-format'),
    (SVGNS, 'font-face-name'),
    (SVGNS, 'stop'),
    (TABLENS, 'body'),
    (TABLENS, 'cell-address'),
    (TABLENS, 'cell-range-source'),
    (TABLENS, 'change-deletion'),
    (TABLENS, 'consolidation'),
    (TABLENS, 'database-source-query'),
    (TABLENS, 'database-source-sql'),
    (TABLENS, 'database-source-table'),
    (TABLENS, 'data-pilot-display-info'),
    (TABLENS, 'data-pilot-field-reference'),
    (TABLENS, 'data-pilot-group-member'),
    (TABLENS, 'data-pilot-layout-info'),
    (TABLENS, 'data-pilot-member'),
    (TABLENS, 'data-pilot-sort-info'),
    (TABLENS, 'data-pilot-subtotal'),
    (TABLENS, 'dependency'),
    (TABLENS, 'error-macro'),
    (TABLENS, 'even-columns'),
    (TABLENS, 'even-rows'),
    (TABLENS, 'filter-condition'),
    (TABLENS, 'first-column'),
    (TABLENS, 'first-row'),
    (TABLENS, 'highlighted-range'),
    (TABLENS, 'insertion-cut-off'),
    (TABLENS, 'iteration'),
    (TABLENS, 'label-range'),
    (TABLENS, 'last-column'),
    (TABLENS, 'last-row'),
    (TABLENS, 'movement-cut-off'),
    (TABLENS, 'named-expression'),
    (TABLENS, 'named-range'),
    (TABLENS, 'null-date'),
    (TABLENS, 'odd-columns'),
    (TABLENS, 'odd-rows'),
    (TABLENS, 'operation'),
    (TABLENS, 'scenario'),
    (TABLENS, 'sort-by'),
    (TABLENS, 'sort-groups'),
    (TABLENS, 'source-range-address'),
    (TABLENS, 'source-service'),
    (TABLENS, 'subtotal-field'),
    (TABLENS, 'table-column'),
    (TABLENS, 'table-source'),
    (TABLENS, 'target-range-address'),
    (TEXTNS, 'alphabetical-index-auto-mark-file'),
    (TEXTNS, 'alphabetical-index-mark'),
    (TEXTNS, 'alphabetical-index-mark-end'),
    (TEXTNS, 'alphabetical-index-mark-start'),
    (TEXTNS, 'bookmark'),
    (TEXTNS, 'bookmark-end'),
    (TEXTNS, 'bookmark-start'),
    (TEXTNS, 'change'),
    (TEXTNS, 'change-end'),
    (TEXTNS, 'change-start'),
    (TEXTNS, 'dde-connection-decl'),
    (TEXTNS, 'index-entry-bibliography'),
    (TEXTNS, 'index-entry-chapter'),
    (TEXTNS, 'index-entry-link-end'),
    (TEXTNS, 'index-entry-link-start'),
    (TEXTNS, 'index-entry-page-number'),
    (TEXTNS, 'index-entry-tab-stop'),
    (TEXTNS, 'index-entry-text'),
    (TEXTNS, 'index-source-style'),
    (TEXTNS, 'line-break'),
    (TEXTNS, 'page'),
    (TEXTNS, 'reference-mark'),
    (TEXTNS, 'reference-mark-end'),
    (TEXTNS, 'reference-mark-start'),
    (TEXTNS, 's'),
    (TEXTNS, 'section-source'),
    (TEXTNS, 'sequence-decl'),
    (TEXTNS, 'soft-page-break'),
    (TEXTNS, 'sort-key'),
    (TEXTNS, 'tab'),
    (TEXTNS, 'toc-mark'),
    (TEXTNS, 'toc-mark-end'),
    (TEXTNS, 'toc-mark-start'),
    (TEXTNS, 'user-field-decl'),
    (TEXTNS, 'user-index-mark'),
    (TEXTNS, 'user-index-mark-end'),
    (TEXTNS, 'user-index-mark-start'),
    (TEXTNS, 'variable-decl'),
)



#!/usr/bin/env python
# License: GPLv3 Copyright: 2012, Kovid Goyal <kovid at kovidgoyal.net>

import re
import traceback

from calibre.constants import iswindows

supernote_settings = {
    'calibre_file_paths': {
        'metadata': ['Document/metadata.calibre', 'Documents/metadata.calibre'],
        'driveinfo': ['Document/driveinfo.calibre', 'Documents/driveinfo.calibre'],
    },
    'send_to': ['Document', 'Documents'],
}


class DeviceDefaults:
    def __init__(self):
        self.rules = (
            # Amazon devices
            (
                {'vendor': 0x1949},
                {
                    'format_map': ['azw3', 'mobi', 'azw', 'azw1', 'azw4', 'kfx', 'pdf'],
                    'send_to': ['documents', 'kindle', 'books'],
                    'apnx': {
                        'send': True,
                        'method': 'fast',
                        'custom_column_page_count': None,
                        'custom_column_method': None,
                    },
                },
            ),
            # B&N devices
            (
                {'vendor': 0x2080},
                {
                    'format_map': ['epub', 'pdf'],
                    # NOOK does not allow writing files into root
                    'calibre_file_paths': {'metadata': 'NOOK/metadata.calibre', 'driveinfo': 'NOOK/driveinfo.calibre'},
                    'send_to': [
                        'NOOK/My Books',
                        'NOOK/My Files',
                        'NOOK',
                        'Calibre_Companion',
                        'Books',
                        'eBooks/import',
                        'eBooks',
                        'sdcard/ebooks',
                    ],
                },
            ),
            # Supernote A5 and A5X and A6X2
            ({'vendor': 0x2207, 'product': 0x0031}, supernote_settings),
            ({'vendor': 0x2207, 'product': 0x0011}, supernote_settings),
            ({'vendor': 0x2207, 'product': 0x0007}, supernote_settings),  # A6X2
            ({'vendor': 0x2207, 'product': 0x0017}, supernote_settings),  # A6X2
        )

    def __call__(self, device, driver):
        if iswindows:
            vid = pid = 0xFFFF
            m = re.search(r'(?i)vid_([0-9a-fA-F]+)&pid_([0-9a-fA-F]+)', device)
            if m is not None:
                try:
                    vid, pid = int(m.group(1), 16), int(m.group(2), 16)
                except Exception:
                    traceback.print_exc()
        else:
            vid, pid = device.vendor_id, device.product_id

        for rule in self.rules:
            tests = rule[0]
            matches = True
            for k, v in tests.items():
                if k == 'vendor' and v != vid:
                    matches = False
                    break
                if k == 'product' and v != pid:
                    matches = False
                    break
            if matches:
                ans = rule[1]
                return ans, vid, pid

        return {}, vid, pid



#!/usr/bin/env python
# License: GPLv3 Copyright: 2026, Kovid Goyal <kovid at kovidgoyal.net>

import os

from qt.core import QDialog

from calibre.constants import ismacos
from calibre.gui2 import error_dialog
from calibre.gui2.actions import InterfaceActionWithLibraryDrop
from calibre.gui2.dialogs.choose_format import ChooseFormatDialog
from calibre.ptempfile import PersistentTemporaryDirectory
from calibre.utils.localization import _


class AirdropAction(InterfaceActionWithLibraryDrop):
    name = 'AirDrop books'
    action_spec = (_('AirDrop books'), 'send.png', _('Share the selected books via AirDrop'), ())
    action_type = 'current'
    dont_add_to = frozenset(('context-menu-device',))

    def genesis(self):
        self.qaction.triggered.connect(self.airdrop_books)

    def do_drop(self):
        book_ids = self.dropped_ids
        del self.dropped_ids
        if book_ids:
            self.airdrop_book_ids(book_ids)

    def location_selected(self, loc):
        self.qaction.setEnabled(loc == 'library')

    def airdrop_books(self):
        book_ids = self.gui.library_view.get_selected_ids()
        if not book_ids:
            return error_dialog(self.gui, _('Cannot AirDrop'), _('No book selected'), show=True)
        self.airdrop_book_ids(book_ids)

    def airdrop_book_ids(self, book_ids):
        if not ismacos:
            return error_dialog(self.gui, _('Cannot AirDrop'), _('AirDrop is only available on macOS'), show=True)
        db = self.gui.current_db.new_api
        formats = {book_id: [f.upper() for f in db.formats(book_id)] for book_id in book_ids}
        all_fmts = {f for fmts in formats.values() for f in fmts}
        if not all_fmts:
            return error_dialog(self.gui, _('Cannot AirDrop'), _('The selected books have no files to share'), show=True)

        def all_have_format(f: str) -> bool:
            for q in formats.values():
                if f not in q:
                    return False
            return True

        common_formats = {f for f in all_fmts if all_have_format(f)}
        if not common_formats:
            return error_dialog(self.gui, _('Cannot AirDrop'), _('The selected books do not all have at least one format in common'), show=True)
        all_fmts = common_formats
        if len(all_fmts) > 1:
            d = ChooseFormatDialog(self.gui, _('Choose the format to AirDrop'), all_fmts)
            if d.exec() != QDialog.DialogCode.Accepted or not d.format():
                return
            fmt = d.format()
        else:
            fmt = tuple(all_fmts)[0]
        temp_dir = PersistentTemporaryDirectory()

        from calibre.gui2.save import Saver

        class Exporter(Saver):
            pd_title = _('Exporting {} books to AirDrop...').format(len(book_ids))

            def on_complete(self):
                paths = tuple(os.path.join(temp_dir, x) for x in os.listdir(temp_dir))
                from calibre_extensions.cocoa import airdrop_share

                try:
                    airdrop_share(*paths)
                except Exception as err:
                    import traceback

                    return error_dialog(self.parent(), _('AirDrop failed'), str(err), det_msg=traceback.format_exc(), show=True)

        self.gui.iactions['Save To Disk'].save_library_format_by_ids(book_ids, fmt, path=temp_dir, saver_class=Exporter)



#!/usr/bin/env python
# License: GPLv3 Copyright: 2011, Kovid Goyal <kovid@kovidgoyal.net>

import numbers
from collections import defaultdict
from collections.abc import Callable, Iterable
from datetime import datetime, timedelta
from typing import cast

from calibre.ebooks.metadata import author_to_author_sort
from calibre.utils.date import UNDEFINED_DATE, parse_date, utc_tz
from calibre.utils.icu import lower as icu_lower
from calibre_extensions.speedup import parse_date as _c_speedup


def identity(x):
    return x


def c_parse(val):
    try:
        year, month, day, hour, minutes, seconds, tzsecs = _c_speedup(val)  # ty: ignore[not-iterable]
    except AttributeError, TypeError:
        # If a value like 2001 is stored in the column, apsw will return it as
        # an int
        if isinstance(val, numbers.Number):
            return datetime(int(val), 1, 3, tzinfo=utc_tz)
        if val is None:
            return UNDEFINED_DATE
    except Exception:
        pass
    else:
        try:
            ans = datetime(year, month, day, hour, minutes, seconds, tzinfo=utc_tz)
            if tzsecs != 0:
                ans -= timedelta(seconds=tzsecs)
        except OverflowError:
            ans = UNDEFINED_DATE
        return ans
    try:
        return parse_date(val, as_utc=True, assume_utc=True)
    except ValueError, TypeError:
        return UNDEFINED_DATE


ONE_ONE, MANY_ONE, MANY_MANY = range(3)

null = object()


class Table:
    supports_notes = False

    def __init__(self, name, metadata, link_table=None):
        self.name, self.metadata = name, metadata
        self.sort_alpha = metadata.get('is_multiple', False) and metadata.get('display', {}).get('sort_alpha', False)
        dt = metadata['datatype']

        # self.unserialize() maps values from the db to python objects
        self.unserialize = {'datetime': c_parse, 'bool': bool}.get(dt)
        self.serialize = None
        if name == 'authors':
            # Legacy
            self.unserialize = lambda x: x.replace('|', ',') if x else ''
            self.serialize = lambda x: x.replace(',', '|')
        self.link_table = link_table or 'books_{}_link'.format(self.metadata['table'])
        if self.supports_notes and dt == 'rating':  # custom ratings table
            self.supports_notes = False

    def remove_books(self, book_ids, db):
        return set()

    def fix_link_table(self, db):
        pass

    def fix_case_duplicates(self, db):
        """If this table contains entries that differ only by case, then merge
        those entries. This can happen in databases created with old versions
        of calibre and non-ascii values, since sqlite's NOCASE only works with
        ascii text."""
        pass


class VirtualTable(Table):
    """
    A dummy table used for fields that only exist in memory like ondevice
    """

    def __init__(self, name, table_type=ONE_ONE, datatype='text'):
        metadata = {'datatype': datatype, 'table': name}
        self.table_type = table_type
        Table.__init__(self, name, metadata)


class OneToOneTable(Table):
    """
    Represents data that is unique per book (it may not actually be unique) but
    each item is assigned to a book in a one-to-one mapping. For example: uuid,
    timestamp, size, etc.
    """

    table_type = ONE_ONE

    def read(self, db):
        idcol = 'id' if self.metadata['table'] == 'books' else 'book'
        query = db.execute('SELECT {}, {} FROM {}'.format(idcol, self.metadata['column'], self.metadata['table']))
        if self.unserialize is None:
            try:
                self.book_col_map = dict(query)
            except UnicodeDecodeError:
                # The db is damaged, try to work around it by ignoring
                # failures to decode utf-8
                query = db.execute('SELECT {}, cast({} as blob) FROM {}'.format(idcol, self.metadata['column'], self.metadata['table']))
                self.book_col_map = {k: bytes(val).decode('utf-8', 'replace') for k, val in query}
        else:
            us = self.unserialize
            self.book_col_map = {book_id: us(val) for book_id, val in query}

    def remove_books(self, book_ids, db):
        clean = set()
        for book_id in book_ids:
            val = self.book_col_map.pop(book_id, null)
            if val is not null:
                clean.add(val)
        return clean


class PathTable(OneToOneTable):
    def set_path(self, book_id, path, db):
        self.book_col_map[book_id] = path
        db.execute('UPDATE books SET path=? WHERE id=?', (path, book_id))


class SizeTable(OneToOneTable):
    def read(self, db):
        query = db.execute('SELECT books.id, (SELECT MAX(uncompressed_size) FROM data WHERE data.book=books.id) FROM books')
        self.book_col_map = dict(query)

    def update_sizes(self, size_map):
        self.book_col_map.update(size_map)


class UUIDTable(OneToOneTable):
    def read(self, db):
        OneToOneTable.read(self, db)
        self.uuid_to_id_map = {v: k for k, v in self.book_col_map.items()}

    def update_uuid_cache(self, book_id_val_map):
        for book_id, uuid in book_id_val_map.items():
            self.uuid_to_id_map.pop(self.book_col_map.get(book_id, None), None)  # discard old uuid
            self.uuid_to_id_map[uuid] = book_id

    def remove_books(self, book_ids, db):
        clean = set()
        for book_id in book_ids:
            val = self.book_col_map.pop(book_id, null)
            if val is not null:
                self.uuid_to_id_map.pop(val, None)
                clean.add(val)
        return clean

    def lookup_by_uuid(self, uuid):
        return self.uuid_to_id_map.get(uuid, None)


class CompositeTable(OneToOneTable):
    def read(self, db):
        self.book_col_map = {}
        d = self.metadata['display']
        self.composite_template = ['composite_template']
        self.contains_html = d.get('contains_html', False)
        self.make_category = d.get('make_category', False)
        self.composite_sort = d.get('composite_sort', False)
        self.use_decorations = d.get('use_decorations', False)

    def remove_books(self, book_ids, db):
        return set()


class ManyToOneTable(Table):
    """
    Represents data where one data item can map to many books, for example:
    series or publisher.

    Each book however has only one value for data of this type.
    """

    table_type = MANY_ONE
    supports_notes = True

    def read(self, db):
        self.id_map = {}
        self.link_map = {}
        self.col_book_map = defaultdict(set)
        self.book_col_map = {}
        self.read_id_maps(db)
        self.read_maps(db)

    def read_id_maps(self, db):
        query = db.execute('SELECT id, {}, link FROM {}'.format(self.metadata['column'], self.metadata['table']))
        us = identity if self.unserialize is None else self.unserialize
        for id_, val, link in query:
            self.id_map[id_] = us(val)
            self.link_map[id_] = link

    def read_maps(self, db):
        cbm = self.col_book_map
        bcm = self.book_col_map
        for book, item_id in db.execute('SELECT book, {} FROM {}'.format(self.metadata['link_column'], self.link_table)):
            cbm[item_id].add(book)
            bcm[book] = item_id

    def fix_link_table(self, db):
        linked_item_ids = set(self.book_col_map.values())
        extra_item_ids = linked_item_ids - set(self.id_map)
        if extra_item_ids:
            for item_id in extra_item_ids:
                book_ids = self.col_book_map.pop(item_id, ())
                for book_id in book_ids:
                    self.book_col_map.pop(book_id, None)
            db.executemany(
                'DELETE FROM {} WHERE {}=?'.format(self.link_table, self.metadata['link_column']),
                tuple((x,) for x in extra_item_ids),
            )

    def fix_case_duplicates(self, db):
        case_map = defaultdict(set)
        for item_id, val in self.id_map.items():
            case_map[icu_lower(val)].add(item_id)

        for v in case_map.values():
            if len(v) > 1:
                main_id = min(v)
                v.discard(main_id)
                item_map = {}
                for item_id in v:
                    val = self.id_map.pop(item_id, null)
                    if val is not null:
                        item_map[item_id] = val
                    books = self.col_book_map.pop(item_id, set())
                    for book_id in books:
                        self.book_col_map[book_id] = main_id
                db.executemany(
                    'UPDATE {0} SET {1}=? WHERE {1}=?'.format(self.link_table, self.metadata['link_column']),
                    tuple((main_id, x) for x in v),
                )
                db.delete_category_items(self.name, self.metadata['table'], item_map)

    def item_ids_for_names(self, db, item_names: Iterable[str], case_sensitive: bool = False) -> dict[str, int | None]:
        item_names = tuple(item_names)
        if case_sensitive:
            colname = self.metadata['column']
            serialized_names = tuple(map(self.serialize, item_names)) if self.serialize else item_names
            if len(item_names) == 1:
                iid = db.get(f'SELECT id FROM {self.metadata["table"]} WHERE {colname} = ?', ((serialized_names[0],)), all=False)
                return {item_names[0]: iid}
            inq = ('?,' * len(item_names))[:-1]
            ans: dict[str, int | None] = {name: None for name in item_names}
            res: Iterable[tuple[str, int]] = db.get(f'SELECT {colname}, id FROM {self.metadata["table"]} WHERE {colname} IN ({inq})', serialized_names)
            if self.unserialize is not None:
                unserialize = cast(Callable[[str], str], self.unserialize)
                res = ((unserialize(name), iid) for name, iid in res)
            ans.update(res)
            return ans
        if len(item_names) == 1:
            q = icu_lower(item_names[0])
            for iid, name in self.id_map.items():
                if icu_lower(name) == q:
                    return {item_names[0]: iid}
            return {item_names[0]: None}
        rmap = {icu_lower(v) if isinstance(v, str) else v: k for k, v in self.id_map.items()}
        return {name: rmap.get(icu_lower(name) if isinstance(name, str) else name, None) for name in item_names}

    def remove_books(self, book_ids, db):
        clean = set()
        for book_id in book_ids:
            item_id = self.book_col_map.pop(book_id, None)
            if item_id is not None:
                try:
                    self.col_book_map[item_id].discard(book_id)
                except KeyError:
                    if self.id_map.pop(item_id, null) is not null:
                        clean.add(item_id)
                else:
                    if not self.col_book_map[item_id]:
                        del self.col_book_map[item_id]
                        if self.id_map.pop(item_id, null) is not null:
                            clean.add(item_id)
        if clean:
            db.executemany('DELETE FROM {} WHERE id=?'.format(self.metadata['table']), [(x,) for x in clean])
        return clean

    def remove_items(self, item_ids, db, restrict_to_book_ids=None):
        affected_books = set()

        if restrict_to_book_ids is not None:
            items_to_process_normally = set()
            # Check if all the books with the item are in the restriction. If
            # so, process them normally
            for item_id in item_ids:
                books_to_process = self.col_book_map.get(item_id, set())
                books_not_to_delete = books_to_process - restrict_to_book_ids
                if books_not_to_delete:
                    # Some books not in restriction. Must do special processing
                    books_to_delete = books_to_process & restrict_to_book_ids
                    # remove the books from the old id maps
                    self.col_book_map[item_id] = books_not_to_delete
                    for book_id in books_to_delete:
                        self.book_col_map.pop(book_id, None)
                    if books_to_delete:
                        # Delete links to the affected books from the link table. As
                        # this is a many-to-one mapping we know that we can delete
                        # links without checking the item ID
                        db.executemany(f'DELETE FROM {self.link_table} WHERE book=?', tuple((x,) for x in books_to_delete))
                        affected_books |= books_to_delete
                else:
                    # Process normally any items where the VL was not significant
                    items_to_process_normally.add(item_id)
            if items_to_process_normally:
                affected_books |= self.remove_items(items_to_process_normally, db)
            return affected_books

        item_map = {}
        for item_id in item_ids:
            val = self.id_map.pop(item_id, null)
            if val is null:
                continue
            item_map[item_id] = val
            book_ids = self.col_book_map.pop(item_id, set())
            for book_id in book_ids:
                self.book_col_map.pop(book_id, None)
            affected_books.update(book_ids)
        db.delete_category_items(self.name, self.metadata['table'], item_map, self.link_table, self.metadata['link_column'])
        return affected_books

    def rename_item(self, item_id, new_name, db):
        existing_item = None
        q = icu_lower(new_name)
        for q_id, q_val in self.id_map.items():
            if icu_lower(q_val) == q:
                existing_item = q_id
                break
        table, col, lcol = self.metadata['table'], self.metadata['column'], self.metadata['link_column']
        affected_books = self.col_book_map.get(item_id, set())
        new_id = item_id
        if existing_item is None or existing_item == item_id:
            # A simple rename will do the trick
            self.id_map[item_id] = new_name
            db.execute(f'UPDATE {table} SET {col}=? WHERE id=?', (new_name, item_id))
        else:
            # We have to replace
            new_id = existing_item
            self.id_map.pop(item_id, None)
            self.link_map.pop(item_id, None)
            books = self.col_book_map.pop(item_id, set())
            for book_id in books:
                self.book_col_map[book_id] = existing_item
            self.col_book_map[existing_item].update(books)
            db.rename_category_item(self.name, table, self.link_table, lcol, item_id, existing_item, self.id_map[new_id])
        return affected_books, new_id

    def set_links(self, link_map, db):
        link_map = {id_: (l or '').strip() for id_, l in link_map.items()}
        link_map = {id_: l for id_, l in link_map.items() if l != self.link_map.get(id_)}
        if link_map:
            self.link_map.update(link_map)
            db.executemany(f'UPDATE {self.metadata["table"]} SET link=? WHERE id=?', tuple((v, k) for k, v in link_map.items()))
        return link_map


class RatingTable(ManyToOneTable):
    supports_notes = False

    def read_id_maps(self, db):
        ManyToOneTable.read_id_maps(self, db)
        # Ensure there are no records with rating=0 in the table. These should
        # be represented as rating:None instead.
        bad_ids = {item_id for item_id, rating in self.id_map.items() if rating == 0}
        if bad_ids:
            self.id_map = {item_id: rating for item_id, rating in self.id_map.items() if rating != 0}
            db.executemany(
                'DELETE FROM {} WHERE {}=?'.format(self.link_table, self.metadata['link_column']),
                tuple((x,) for x in bad_ids),
            )
            db.execute('DELETE FROM {} WHERE {}=0'.format(self.metadata['table'], self.metadata['column']))


class ManyToManyTable(ManyToOneTable):
    """
    Represents data that has a many-to-many mapping with books. i.e. each book
    can have more than one value and each value can be mapped to more than one
    book. For example: tags or authors.
    """

    table_type = MANY_MANY
    selectq = 'SELECT book, {0} FROM {1} ORDER BY id'
    do_clean_on_remove = True

    def read_maps(self, db):
        bcm = defaultdict(list)
        cbm = self.col_book_map
        for book, item_id in db.execute(self.selectq.format(self.metadata['link_column'], self.link_table)):
            cbm[item_id].add(book)
            bcm[book].append(item_id)

        self.book_col_map = {k: tuple(v) for k, v in bcm.items()}

    def fix_link_table(self, db):
        linked_item_ids = {item_id for item_ids in self.book_col_map.values() for item_id in item_ids}
        extra_item_ids = linked_item_ids - set(self.id_map)
        if extra_item_ids:
            for item_id in extra_item_ids:
                book_ids = self.col_book_map.pop(item_id, ())
                for book_id in book_ids:
                    self.book_col_map[book_id] = tuple(iid for iid in self.book_col_map.pop(book_id, ()) if iid not in extra_item_ids)
            db.executemany(
                'DELETE FROM {} WHERE {}=?'.format(self.link_table, self.metadata['link_column']),
                tuple((x,) for x in extra_item_ids),
            )

    def remove_books(self, book_ids, db):
        clean = {}
        for book_id in book_ids:
            item_ids = self.book_col_map.pop(book_id, ())
            for item_id in item_ids:
                try:
                    self.col_book_map[item_id].discard(book_id)
                except KeyError:
                    if (val := self.id_map.pop(item_id, null)) is not null:
                        clean[item_id] = val
                else:
                    if not self.col_book_map[item_id]:
                        del self.col_book_map[item_id]
                        val = self.id_map.pop(item_id, null)
                        if val is not null:
                            clean[item_id] = val
        if clean and self.do_clean_on_remove:
            db.delete_category_items(self.name, self.metadata['table'], clean)
        return set(clean)

    def remove_items(self, item_ids, db, restrict_to_book_ids=None):
        affected_books = set()
        if restrict_to_book_ids is not None:
            items_to_process_normally = set()
            # Check if all the books with the item are in the restriction. If
            # so, process them normally
            for item_id in item_ids:
                books_to_process = self.col_book_map.get(item_id, set())
                books_not_to_delete = books_to_process - restrict_to_book_ids
                if books_not_to_delete:
                    # Some books not in restriction. Must do special processing
                    books_to_delete = books_to_process & restrict_to_book_ids
                    # remove the books from the old id maps
                    self.col_book_map[item_id] = books_not_to_delete
                    for book_id in books_to_delete:
                        self.book_col_map[book_id] = tuple(x for x in self.book_col_map.get(book_id, ()) if x != item_id)
                    affected_books |= books_to_delete
                else:
                    items_to_process_normally.add(item_id)
            # Delete book/item pairs from the link table. We don't need to do
            # anything with the main table because books with the old ID are
            # still in the library.
            db.executemany(
                'DELETE FROM {} WHERE {}=? and {}=?'.format(self.link_table, 'book', self.metadata['link_column']),
                [(b, i) for b in affected_books for i in item_ids],
            )
            # Take care of any items where the VL was not significant
            if items_to_process_normally:
                affected_books |= self.remove_items(items_to_process_normally, db)
            return affected_books

        item_map = {}
        for item_id in item_ids:
            val = self.id_map.pop(item_id, null)
            if val is null:
                continue
            item_map[item_id] = val
            book_ids = self.col_book_map.pop(item_id, set())
            for book_id in book_ids:
                self.book_col_map[book_id] = tuple(x for x in self.book_col_map.get(book_id, ()) if x != item_id)
            affected_books.update(book_ids)
        db.delete_category_items(self.name, self.metadata['table'], item_map, self.link_table, self.metadata['link_column'])
        return affected_books

    def rename_item(self, item_id, new_name, db):
        existing_item = None
        q = icu_lower(new_name)
        for q_id, q_val in self.id_map.items():
            if icu_lower(q_val) == q:
                existing_item = q_id
                break
        table, col, lcol = self.metadata['table'], self.metadata['column'], self.metadata['link_column']
        affected_books = self.col_book_map.get(item_id, set())
        new_id = item_id
        if existing_item is None or existing_item == item_id:
            # A simple rename will do the trick
            self.id_map[item_id] = new_name
            nn = self.serialize(new_name) if self.serialize else new_name
            db.execute(f'UPDATE {table} SET {col}=? WHERE id=?', (nn, item_id))
        else:
            # We have to replace
            new_id = existing_item
            self.id_map.pop(item_id, None)
            self.link_map.pop(item_id, None)
            books = self.col_book_map.pop(item_id, set())
            # Replacing item_id with existing_item could cause the same id to
            # appear twice in the book list. Handle that by removing existing
            # item from the book list before replacing.
            for book_id in books:
                self.book_col_map[book_id] = tuple((existing_item if x == item_id else x) for x in self.book_col_map.get(book_id, ()) if x != existing_item)
            self.col_book_map[existing_item].update(books)
            db.executemany(
                f'DELETE FROM {self.link_table} WHERE book=? AND {lcol}=?',
                [(book_id, existing_item) for book_id in books],
            )
            db.rename_category_item(self.name, table, self.link_table, lcol, item_id, existing_item, self.id_map[new_id])
        return affected_books, new_id

    def fix_case_duplicates(self, db):
        from calibre.db.write import uniq

        case_map = defaultdict(set)
        for item_id, val in self.id_map.items():
            case_map[icu_lower(val)].add(item_id)

        for v in case_map.values():
            if len(v) > 1:
                done_books = set()
                main_id = min(v)
                v.discard(main_id)
                item_map = {}
                for item_id in v:
                    val = self.id_map.pop(item_id, null)
                    if val is not null:
                        item_map[item_id] = val
                    books = self.col_book_map.pop(item_id, set())
                    for book_id in books:
                        if book_id in done_books:
                            continue
                        done_books.add(book_id)
                        orig = self.book_col_map.get(book_id, ())
                        if not orig:
                            continue
                        vals = uniq(tuple(main_id if x in v else x for x in orig))
                        self.book_col_map[book_id] = vals
                        if len(orig) == len(vals):
                            # We have a simple replacement
                            db.executemany(
                                'UPDATE {0} SET {1}=? WHERE {1}=? AND book=?'.format(self.link_table, self.metadata['link_column']),
                                tuple((main_id, x, book_id) for x in v),
                            )
                        else:
                            # duplicates
                            db.execute(f'DELETE FROM {self.link_table} WHERE book=?', (book_id,))
                            db.executemany(
                                'INSERT INTO {} (book,{}) VALUES (?,?)'.format(self.link_table, self.metadata['link_column']),
                                tuple((book_id, x) for x in vals),
                            )
                db.delete_category_items(self.name, self.metadata['table'], item_map)


class AuthorsTable(ManyToManyTable):
    def read_id_maps(self, db):
        self.link_map = lm = {}
        self.asort_map = sm = {}
        self.id_map = im = {}
        us = self.unserialize
        assert us is not None
        for aid, name, sort, link in db.execute('SELECT id, name, sort, link FROM authors'):
            name = us(name)
            im[aid] = name
            sm[aid] = sort or author_to_author_sort(name)
            lm[aid] = link

    def set_sort_names(self, aus_map, db):
        aus_map = {aid: (a or '').strip() for aid, a in aus_map.items()}
        aus_map = {aid: a for aid, a in aus_map.items() if a != self.asort_map.get(aid, None)}
        self.asort_map.update(aus_map)
        db.executemany('UPDATE authors SET sort=? WHERE id=?', [(v, k) for k, v in aus_map.items()])
        return aus_map

    def set_links(self, link_map, db):
        link_map = {aid: (l or '').strip() for aid, l in link_map.items()}
        link_map = {aid: l for aid, l in link_map.items() if l != self.link_map.get(aid, None)}
        self.link_map.update(link_map)
        db.executemany('UPDATE authors SET link=? WHERE id=?', [(v, k) for k, v in link_map.items()])
        return link_map

    def remove_books(self, book_ids, db):
        clean = ManyToManyTable.remove_books(self, book_ids, db)
        for item_id in clean:
            self.link_map.pop(item_id, None)
            self.asort_map.pop(item_id, None)
        return clean

    def rename_item(self, item_id, new_name, db):
        ret = ManyToManyTable.rename_item(self, item_id, new_name, db)
        if item_id not in self.id_map:
            self.asort_map.pop(item_id, None)
        else:
            # Was a simple rename, update the author sort value
            self.set_sort_names({item_id: author_to_author_sort(new_name)}, db)

        return ret

    def remove_items(self, item_ids, db, restrict_to_book_ids=None):
        raise NotImplementedError('Direct removal of authors is not allowed')


class FormatsTable(ManyToManyTable):
    do_clean_on_remove = False
    supports_notes = False

    def read_id_maps(self, db):
        pass

    def fix_case_duplicates(self, db):
        pass

    def read_maps(self, db):
        self.fname_map = fnm = defaultdict(dict)
        self.size_map = sm = defaultdict(dict)
        self.col_book_map = cbm = defaultdict(set)
        bcm = defaultdict(list)

        for book, fmt, name, sz in db.execute('SELECT book, format, name, uncompressed_size FROM data'):
            if fmt is not None:
                fmt = fmt.upper()
                cbm[fmt].add(book)
                bcm[book].append(fmt)
                fnm[book][fmt] = name
                sm[book][fmt] = sz

        self.book_col_map = {k: tuple(sorted(v)) for k, v in bcm.items()}

    def remove_books(self, book_ids, db):
        clean = ManyToManyTable.remove_books(self, book_ids, db)
        for book_id in book_ids:
            self.fname_map.pop(book_id, None)
            self.size_map.pop(book_id, None)
        return clean

    def set_fname(self, book_id, fmt, fname, db):
        self.fname_map[book_id][fmt] = fname
        db.execute('UPDATE data SET name=? WHERE book=? AND format=?', (fname, book_id, fmt))

    def remove_formats(self, formats_map, db):
        for book_id, fmts in formats_map.items():
            self.book_col_map[book_id] = tuple(fmt for fmt in self.book_col_map.get(book_id, ()) if fmt not in fmts)
            for m in (self.fname_map, self.size_map):
                m[book_id] = {k: v for k, v in m[book_id].items() if k not in fmts}
            for fmt in fmts:
                try:
                    self.col_book_map[fmt].discard(book_id)
                except KeyError:
                    pass
        db.executemany(
            'DELETE FROM data WHERE book=? AND format=?',
            [(book_id, fmt) for book_id, fmts in formats_map.items() for fmt in fmts],
        )

        def zero_max(book_id):
            try:
                return max(self.size_map[book_id].values())
            except ValueError:
                return 0

        return {book_id: zero_max(book_id) for book_id in formats_map}

    def remove_items(self, item_ids, db, restrict_to_book_ids=None):
        raise NotImplementedError('Cannot delete a format directly')

    def rename_item(self, item_id, new_name, db):
        raise NotImplementedError('Cannot rename formats')

    def update_fmt(self, book_id, fmt, fname, size, db):
        fmts = list(self.book_col_map.get(book_id, []))
        try:
            fmts.remove(fmt)
        except ValueError:
            pass
        fmts.append(fmt)
        self.book_col_map[book_id] = tuple(fmts)

        try:
            self.col_book_map[fmt].add(book_id)
        except KeyError:
            self.col_book_map[fmt] = {book_id}

        self.fname_map[book_id][fmt] = fname
        self.size_map[book_id][fmt] = size
        db.execute(
            'INSERT OR REPLACE INTO data (book,format,uncompressed_size,name) VALUES (?,?,?,?)',
            (book_id, fmt, size, fname),
        )
        return max(self.size_map[book_id].values())


class IdentifiersTable(ManyToManyTable):
    supports_notes = False

    def read_id_maps(self, db):
        pass

    def fix_case_duplicates(self, db):
        pass

    def read_maps(self, db):
        self.book_col_map = defaultdict(dict)
        self.col_book_map = defaultdict(set)
        for book, typ, val in db.execute('SELECT book, type, val FROM identifiers'):
            if typ is not None and val is not None:
                self.col_book_map[typ].add(book)
                self.book_col_map[book][typ] = val

    def remove_books(self, book_ids, db):
        clean = set()
        for book_id in book_ids:
            item_map = self.book_col_map.pop(book_id, {})
            for item_id in item_map:
                try:
                    self.col_book_map[item_id].discard(book_id)
                except KeyError:
                    clean.add(item_id)
                else:
                    if not self.col_book_map[item_id]:
                        del self.col_book_map[item_id]
                        clean.add(item_id)
        return clean

    def remove_items(self, item_ids, db, restrict_to_book_ids=None):
        raise NotImplementedError('Direct deletion of identifiers is not implemented')

    def rename_item(self, item_id, new_name, db):
        raise NotImplementedError('Cannot rename identifiers')

    def all_identifier_types(self):
        return frozenset(k for k, v in self.col_book_map.items() if v)



#!/usr/bin/env python
# License: GPLv3 Copyright: 2018, Kovid Goyal <kovid at kovidgoyal.net>

from qt.core import QAction, Qt, QToolButton

from calibre.gui2.actions import InterfaceAction
from calibre.utils.localization import _


class VirtualLibraryAction(InterfaceAction):
    name = 'Virtual library'
    action_spec = (_('Virtual library'), 'vl.png', _('Change the current Virtual library'), None)
    action_type = 'current'
    action_add_menu = True
    popup_type = QToolButton.ToolButtonPopupMode.InstantPopup
    dont_add_to = frozenset(('context-menu-device', 'menubar-device'))

    def genesis(self):
        self.menu = m = self.qaction.menu()
        assert m is not None
        m.aboutToShow.connect(self.about_to_show_menu)
        self.qs_action = QAction(self.gui)
        self.gui.addAction(self.qs_action)
        self.qs_action.triggered.connect(self.gui.choose_vl_triggerred)
        self.gui.keyboard.register_shortcut(
            self.unique_name + ' - ' + 'quick-select-vl',
            _('Quick select Virtual library'),
            default_keys=('Ctrl+T',),
            action=self.qs_action,
            description=_('Quick select a Virtual library'),
            group=self.action_spec[0],
        )
        ac = self.create_action(
            spec=(_('Switch to last used Virtual library'), 'vl.png', None, None),
            attr='action_previous_virtual_library',
        )
        ac.triggered.connect(self.switch_to_previous_virtual_library, type=Qt.ConnectionType.QueuedConnection)
        self.gui.keyboard.register_shortcut(
            self.unique_name + '-' + 'action_previous_virtual_library',
            ac.text(),
            action=ac,
            group=self.action_spec[0],
            default_keys=('Ctrl+Alt+Shift+P',),
        )
        self.gui.addAction(ac)

    def about_to_show_menu(self):
        self.gui.build_virtual_library_menu(self.menu, add_tabs_action=False)

    def library_changed(self, db):
        self.gui.clear_vl_history()

    def switch_to_previous_virtual_library(self):
        self.gui.switch_to_previous_virtual_library()



#!/usr/bin/env python
# License: GPLv3 Copyright: 2013, Kovid Goyal <kovid at kovidgoyal.net>

from calibre.utils.run_tests import find_tests_in_package, run_tests


def find_tests():
    return find_tests_in_package('calibre.db.tests')


if __name__ == '__main__':
    try:
        import init_calibre  # noqa: F401
    except ImportError:
        pass
    run_tests(find_tests)



#!/usr/bin/env python

__author__ = 'stackoverflow community'
'''
Takes english numeric words and converts them to integers.
Returns False if the word isn't a number.

implementation courtesy of the stackoverflow community:
http://stackoverflow.com/questions/493174/is-there-a-way-to-convert-number-words-to-integers-python
'''

import re

numwords = {}


def text2int(textnum):
    if not numwords:
        units = [
            'zero',
            'one',
            'two',
            'three',
            'four',
            'five',
            'six',
            'seven',
            'eight',
            'nine',
            'ten',
            'eleven',
            'twelve',
            'thirteen',
            'fourteen',
            'fifteen',
            'sixteen',
            'seventeen',
            'eighteen',
            'nineteen',
        ]

        tens = ['', '', 'twenty', 'thirty', 'forty', 'fifty', 'sixty', 'seventy', 'eighty', 'ninety']

        scales = [
            'hundred',
            'thousand',
            'million',
            'billion',
            'trillion',
            'quadrillion',
            'quintillion',
            'sexillion',
            'septillion',
            'octillion',
            'nonillion',
            'decillion',
        ]

        numwords['and'] = (1, 0)
        for idx, word in enumerate(units):
            numwords[word] = (1, idx)
        for idx, word in enumerate(tens):
            numwords[word] = (1, idx * 10)
        for idx, word in enumerate(scales):
            numwords[word] = (10 ** (idx * 3 or 2), 0)

    ordinal_words = {'first': 1, 'second': 2, 'third': 3, 'fifth': 5, 'eighth': 8, 'ninth': 9, 'twelfth': 12}
    ordinal_endings = [('ieth', 'y'), ('th', '')]
    current = result = 0
    tokens = re.split(r'[\s-]+', textnum)
    for word in tokens:
        if word in ordinal_words:
            scale, increment = (1, ordinal_words[word])
        else:
            for ending, replacement in ordinal_endings:
                if word.endswith(ending):
                    word = f'{word[: -len(ending)]}{replacement}'

            if word not in numwords:
                # raise Exception("Illegal word: " + word)
                return False

            scale, increment = numwords[word]

        if scale > 1:
            current = max(1, current)

        current = current * scale + increment
        if scale > 100:
            result += current
            current = 0

    return result + current



# License: GPLv3 Copyright: 2009, John Schember <john@nachtimwald.com>

"""
Device driver for IRex Iliad
"""

from calibre.devices.usbms.driver import USBMS
from calibre.utils.localization import _


class ILIAD(USBMS):
    name = 'IRex Iliad Device Interface'
    description = _('Communicate with the IRex Iliad e-book reader.')
    author = 'John Schember'
    supported_platforms = ['windows', 'osx', 'linux']

    # Ordered list of supported formats
    # Be sure these have an entry in calibre.devices.mime
    FORMATS = ['mobi', 'prc', 'html', 'pdf', 'txt']

    VENDOR_ID = [0x04CC]
    PRODUCT_ID = [0x1A64]
    BCD = [0x100]

    VENDOR_NAME = 'IREX'
    WINDOWS_MAIN_MEM = 'ILIAD'

    # OSX_MAIN_MEM = ''

    MAIN_MEMORY_VOLUME_LABEL = 'IRex Iliad Main Memory'

    EBOOK_DIR_MAIN = 'books'
    DELETE_EXTS = ['.mbp']
    SUPPORTS_SUB_DIRS = True



#!/usr/bin/env python
# License: GPLv3 Copyright: 2013, Kovid Goyal <kovid at kovidgoyal.net>

from calibre import prepare_string_for_xml
from calibre.gui2.tweak_book import current_container, tprefs

DEFAULT_TEMPLATES = {
    'html': '''\
<?xml version='1.0' encoding='utf-8'?>
<html xmlns="http://www.w3.org/1999/xhtml">
<head>
    <title>{TITLE}</title>
</head>

<body>
    %CURSOR%
</body>
</html>
''',
    'css': '''\
@charset "utf-8";
/* Styles for {TITLE} */
%CURSOR%
''',
}


def raw_template_for(syntax):
    return tprefs['templates'].get(syntax, DEFAULT_TEMPLATES.get(syntax, ''))


def template_for(syntax):
    mi = current_container().mi
    data = {
        'TITLE': mi.title,
        'AUTHOR': ' & '.join(mi.authors),
    }
    return raw_template_for(syntax).format(**{k: prepare_string_for_xml(v, True) for k, v in data.items()})



#########################################################################
#                                                                       #
#                                                                       #
#   copyright 2002 Paul Henry Tremblay                                  #
#                                                                       #
#   This program is distributed in the hope that it will be useful,     #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of      #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU    #
#   General Public License for more details.                            #
#                                                                       #
#                                                                       #
#########################################################################
import os
import re

from calibre.ebooks.rtf2xml import check_brackets, copy
from calibre.ptempfile import better_mktemp

from . import open_for_read, open_for_write


class ProcessTokens:
    '''
    Process each token on a line and add information that will be useful for
    later processing. Information will be put on one line, delimited by "<"
    for main fields, and ">" for sub fields
    '''

    def __init__(
        self,
        in_file,
        exception_handler,
        bug_handler,
        copy=None,
        run_level=1,
    ):
        self.__file = in_file
        self.__bug_handler = bug_handler
        self.__copy = copy
        self.__run_level = run_level
        self.__write_to = better_mktemp()
        self.initiate_token_dict()
        # self.initiate_token_actions()
        self.compile_expressions()
        self.__bracket_count = 0
        self.__exception_handler = exception_handler
        self.__bug_handler = bug_handler

    def compile_expressions(self):
        self.__num_exp = re.compile(r'([a-zA-Z]+)(.*)')
        self.__utf_exp = re.compile(r'(&.*?;)')

    def initiate_token_dict(self):
        self.__return_code = 0
        self.dict_token = {
            # unicode
            'mshex': ('nu', '__________', self.__ms_hex_func),
            # brackets
            '{': ('nu', '{', self.ob_func),
            '}': ('nu', '}', self.cb_func),
            # microsoft characters
            'ldblquote': ('mc', 'ldblquote', self.ms_sub_func),
            'rdblquote': ('mc', 'rdblquote', self.ms_sub_func),
            'rquote': ('mc', 'rquote', self.ms_sub_func),
            'lquote': ('mc', 'lquote', self.ms_sub_func),
            'emdash': ('mc', 'emdash', self.ms_sub_func),
            'endash': ('mc', 'endash', self.ms_sub_func),
            'bullet': ('mc', 'bullet', self.ms_sub_func),
            '~': ('mc', '~', self.ms_sub_func),
            'tab': ('mc', 'tab', self.ms_sub_func),
            '_': ('mc', '_', self.ms_sub_func),
            ';': ('mc', ';', self.ms_sub_func),
            # this must be wrong
            '-': ('mc', '-', self.ms_sub_func),
            'line': ('mi', 'hardline-break', self.direct_conv_func),  # calibre
            # misc => ml
            '*': ('ml', 'asterisk__', self.default_func),
            ':': ('ml', 'colon_____', self.default_func),
            # text
            'backslash': ('nu', '\\', self.text_func),
            'ob': ('nu', '{', self.text_func),
            'cb': ('nu', '}', self.text_func),
            # paragraph formatting => pf
            'page': ('pf', 'page-break', self.default_func),
            'par': ('pf', 'par-end___', self.default_func),
            'pard': ('pf', 'par-def___', self.default_func),
            'keepn': ('pf', 'keep-w-nex', self.bool_st_func),
            'widctlpar': ('pf', 'widow-cntl', self.bool_st_func),
            'adjustright': ('pf', 'adjust-rgt', self.bool_st_func),
            'lang': ('pf', 'language__', self.__language_func),
            'ri': ('pf', 'right-inde', self.divide_by_20),
            'fi': ('pf', 'fir-ln-ind', self.divide_by_20),
            'li': ('pf', 'left-inden', self.divide_by_20),
            'sb': ('pf', 'space-befo', self.divide_by_20),
            'sa': ('pf', 'space-afte', self.divide_by_20),
            'sl': ('pf', 'line-space', self.divide_by_20),
            'deftab': ('pf', 'default-ta', self.divide_by_20),
            'ql': ('pf', 'align_____<left', self.two_part_func),
            'qc': ('pf', 'align_____<cent', self.two_part_func),
            'qj': ('pf', 'align_____<just', self.two_part_func),
            'qr': ('pf', 'align_____<right', self.two_part_func),
            'nowidctlpar': ('pf', 'widow-cntr<false', self.two_part_func),
            'tx': ('pf', 'tab-stop__', self.divide_by_20),
            'tb': ('pf', 'tab-bar-st', self.divide_by_20),
            'tqr': ('pf', 'tab-right_', self.default_func),
            'tqdec': ('pf', 'tab-dec___', self.default_func),
            'tqc': ('pf', 'tab-center', self.default_func),
            'tlul': ('pf', 'leader-und', self.default_func),
            'tlhyph': ('pf', 'leader-hyp', self.default_func),
            'tldot': ('pf', 'leader-dot', self.default_func),
            # stylesheet = > ss
            'stylesheet': ('ss', 'style-shet', self.default_func),
            'sbasedon': ('ss', 'based-on__', self.default_func),
            'snext': ('ss', 'next-style', self.default_func),
            'cs': ('ss', 'char-style', self.default_func),
            's': ('ss', 'para-style', self.default_func),
            # graphics => gr
            'pict': ('gr', 'picture___', self.default_func),
            'objclass': ('gr', 'obj-class_', self.default_func),
            'macpict': ('gr', 'mac-pic___', self.default_func),
            # section => sc
            'sect': ('sc', 'section___', self.default_func),
            'sectd': ('sc', 'sect-defin', self.default_func),
            'endhere': ('sc', 'sect-note_', self.default_func),
            # list=> ls
            'pntext': ('ls', 'list-text_', self.default_func),
            # this line must be wrong because it duplicates an earlier one
            'listtext': ('ls', 'list-text_', self.default_func),
            'pn': ('ls', 'list______', self.default_func),
            'pnseclvl': ('ls', 'list-level', self.default_func),
            'pncard': ('ls', 'list-cardi', self.bool_st_func),
            'pndec': ('ls', 'list-decim', self.bool_st_func),
            'pnucltr': ('ls', 'list-up-al', self.bool_st_func),
            'pnucrm': ('ls', 'list-up-ro', self.bool_st_func),
            'pnord': ('ls', 'list-ord__', self.bool_st_func),
            'pnordt': ('ls', 'list-ordte', self.bool_st_func),
            'pnlvlblt': ('ls', 'list-bulli', self.bool_st_func),
            'pnlvlbody': ('ls', 'list-simpi', self.bool_st_func),
            'pnlvlcont': ('ls', 'list-conti', self.bool_st_func),
            'pnhang': ('ls', 'list-hang_', self.bool_st_func),
            'pntxtb': ('ls', 'list-tebef', self.bool_st_func),
            'ilvl': ('ls', 'list-level', self.default_func),
            'ls': ('ls', 'list-id___', self.default_func),
            'pnstart': ('ls', 'list-start', self.default_func),
            'itap': ('ls', 'nest-level', self.default_func),
            'leveltext': ('ls', 'level-text', self.default_func),
            'levelnumbers': ('ls', 'level-numb', self.default_func),
            'list': ('ls', 'list-in-tb', self.default_func),
            'listlevel': ('ls', 'list-tb-le', self.default_func),
            'listname': ('ls', 'list-name_', self.default_func),
            'listtemplateid': ('ls', 'ls-tem-id_', self.default_func),
            'leveltemplateid': ('ls', 'lv-tem-id_', self.default_func),
            'listhybrid': ('ls', 'list-hybri', self.default_func),
            'levelstartat': ('ls', 'level-star', self.default_func),
            'levelspace': ('ls', 'level-spac', self.divide_by_20),
            'levelindent': ('ls', 'level-inde', self.default_func),
            'levelnfc': ('ls', 'level-type', self.__list_type_func),
            'levelnfcn': ('ls', 'level-type', self.__list_type_func),
            'listid': ('ls', 'lis-tbl-id', self.default_func),
            'listoverride': ('ls', 'lis-overid', self.default_func),
            # duplicate
            'pnlvl': ('ls', 'list-level', self.default_func),
            # root info => ri
            'rtf': ('ri', 'rtf_______', self.default_func),
            'deff': ('ri', 'deflt-font', self.default_func),
            'mac': ('ri', 'macintosh_', self.default_func),
            'pc': ('ri', 'pc________', self.default_func),
            'pca': ('ri', 'pca_______', self.default_func),
            'ansi': ('ri', 'ansi______', self.default_func),
            'ansicpg': ('ri', 'ansi-codpg', self.default_func),
            # notes => nt
            'footnote': ('nt', 'footnote__', self.default_func),
            'ftnalt': ('nt', 'type______<endnote', self.two_part_func),
            # anchor => an
            'tc': ('an', 'toc_______', self.default_func),
            'bkmkstt': ('an', 'book-mk-st', self.default_func),
            'bkmkstart': ('an', 'book-mk-st', self.default_func),
            'bkmkend': ('an', 'book-mk-en', self.default_func),
            'xe': ('an', 'index-mark', self.default_func),
            'rxe': ('an', 'place_____', self.default_func),
            # index => in
            'bxe': ('in', 'index-bold', self.default_func),
            'ixe': ('in', 'index-ital', self.default_func),
            'txe': ('in', 'index-see_', self.default_func),
            # table of contents => tc
            'tcl': ('tc', 'toc-level_', self.default_func),
            'tcn': ('tc', 'toc-sup-nu', self.default_func),
            # field => fd
            'field': ('fd', 'field_____', self.default_func),
            'fldinst': ('fd', 'field-inst', self.default_func),
            'fldrslt': ('fd', 'field-rslt', self.default_func),
            'datafield': ('fd', 'datafield_', self.default_func),
            # info-tables => it
            'fonttbl': ('it', 'font-table', self.default_func),
            'colortbl': ('it', 'colr-table', self.default_func),
            'listoverridetable': ('it', 'lovr-table', self.default_func),
            'listtable': ('it', 'listtable_', self.default_func),
            'revtbl': ('it', 'revi-table', self.default_func),
            # character info => ci
            'b': ('ci', 'bold______', self.bool_st_func),
            'blue': ('ci', 'blue______', self.color_func),
            'caps': ('ci', 'caps______', self.bool_st_func),
            'cf': ('ci', 'font-color', self.colorz_func),
            'chftn': ('ci', 'footnot-mk', self.bool_st_func),
            'dn': ('ci', 'font-down_', self.divide_by_2),
            'embo': ('ci', 'emboss____', self.bool_st_func),
            'f': ('ci', 'font-style', self.default_func),
            'fs': ('ci', 'font-size_', self.divide_by_2),
            'green': ('ci', 'green_____', self.color_func),
            'i': ('ci', 'italics___', self.bool_st_func),
            'impr': ('ci', 'engrave___', self.bool_st_func),
            'outl': ('ci', 'outline___', self.bool_st_func),
            'plain': ('ci', 'plain_____', self.bool_st_func),
            'red': ('ci', 'red_______', self.color_func),
            'scaps': ('ci', 'small-caps', self.bool_st_func),
            'shad': ('ci', 'shadow____', self.bool_st_func),
            'strike': ('ci', 'strike-thr', self.bool_st_func),
            'striked': ('ci', 'dbl-strike', self.bool_st_func),
            'sub': ('ci', 'subscript_', self.bool_st_func),
            'super': ('ci', 'superscrip', self.bool_st_func),
            'nosupersub': ('ci', 'no-su-supe', self.__no_sup_sub_func),
            'up': ('ci', 'font-up___', self.divide_by_2),
            'v': ('ci', 'hidden____', self.default_func),
            # underline
            # can't see why it isn't a char info: 'ul'=>'ci'
            'ul': ('ci', 'underlined<continous', self.two_part_func),
            'uld': ('ci', 'underlined<dotted', self.two_part_func),
            'uldash': ('ci', 'underlined<dash', self.two_part_func),
            'uldashd': ('ci', 'underlined<dash-dot', self.two_part_func),
            'uldashdd': ('ci', 'underlined<dash-dot-dot', self.two_part_func),
            'uldb': ('ci', 'underlined<double', self.two_part_func),
            'ulhwave': ('ci', 'underlined<heavy-wave', self.two_part_func),
            'ulldash': ('ci', 'underlined<long-dash', self.two_part_func),
            'ulth': ('ci', 'underlined<thich', self.two_part_func),
            'ulthd': ('ci', 'underlined<thick-dotted', self.two_part_func),
            'ulthdash': ('ci', 'underlined<thick-dash', self.two_part_func),
            'ulthdashd': ('ci', 'underlined<thick-dash-dot', self.two_part_func),
            'ulthdashdd': ('ci', 'underlined<thick-dash-dot-dot', self.two_part_func),
            'ulthldash': ('ci', 'underlined<thick-long-dash', self.two_part_func),
            'ululdbwave': ('ci', 'underlined<double-wave', self.two_part_func),
            'ulw': ('ci', 'underlined<word', self.two_part_func),
            'ulwave': ('ci', 'underlined<wave', self.two_part_func),
            'ulnone': ('ci', 'underlined<false', self.two_part_func),
            # table => tb
            'trowd': ('tb', 'row-def___', self.default_func),
            'cell': ('tb', 'cell______', self.default_func),
            'row': ('tb', 'row_______', self.default_func),
            'intbl': ('tb', 'in-table__', self.default_func),
            'cols': ('tb', 'columns___', self.default_func),
            'trleft': ('tb', 'row-pos-le', self.divide_by_20),
            'cellx': ('tb', 'cell-posit', self.divide_by_20),
            'trhdr': ('tb', 'row-header', self.default_func),
            # preamble => pr
            # document information => di
            # TODO integrate \userprops
            'info': ('di', 'doc-info__', self.default_func),
            'title': ('di', 'title_____', self.default_func),
            'author': ('di', 'author____', self.default_func),
            'operator': ('di', 'operator__', self.default_func),
            'manager': ('di', 'manager___', self.default_func),
            'company': ('di', 'company___', self.default_func),
            'keywords': ('di', 'keywords__', self.default_func),
            'category': ('di', 'category__', self.default_func),
            'doccomm': ('di', 'doc-notes_', self.default_func),
            'comment': ('di', 'doc-notes_', self.default_func),
            'subject': ('di', 'subject___', self.default_func),
            'creatim': ('di', 'create-tim', self.default_func),
            'yr': ('di', 'year______', self.default_func),
            'mo': ('di', 'month_____', self.default_func),
            'dy': ('di', 'day_______', self.default_func),
            'min': ('di', 'minute____', self.default_func),
            'sec': ('di', 'second____', self.default_func),
            'revtim': ('di', 'revis-time', self.default_func),
            'edmins': ('di', 'edit-time_', self.default_func),
            'printim': ('di', 'print-time', self.default_func),
            'buptim': ('di', 'backuptime', self.default_func),
            'nofwords': ('di', 'num-of-wor', self.default_func),
            'nofchars': ('di', 'num-of-chr', self.default_func),
            'nofcharsws': ('di', 'numofchrws', self.default_func),
            'nofpages': ('di', 'num-of-pag', self.default_func),
            'version': ('di', 'version___', self.default_func),
            'vern': ('di', 'intern-ver', self.default_func),
            'hlinkbase': ('di', 'linkbase__', self.default_func),
            'id': ('di', 'internalID', self.default_func),
            # headers and footers => hf
            'headerf': ('hf', 'head-first', self.default_func),
            'headerl': ('hf', 'head-left_', self.default_func),
            'headerr': ('hf', 'head-right', self.default_func),
            'footerf': ('hf', 'foot-first', self.default_func),
            'footerl': ('hf', 'foot-left_', self.default_func),
            'footerr': ('hf', 'foot-right', self.default_func),
            'header': ('hf', 'header____', self.default_func),
            'footer': ('hf', 'footer____', self.default_func),
            # page => pa
            'margl': ('pa', 'margin-lef', self.divide_by_20),
            'margr': ('pa', 'margin-rig', self.divide_by_20),
            'margb': ('pa', 'margin-bot', self.divide_by_20),
            'margt': ('pa', 'margin-top', self.divide_by_20),
            'gutter': ('pa', 'gutter____', self.divide_by_20),
            'paperw': ('pa', 'paper-widt', self.divide_by_20),
            'paperh': ('pa', 'paper-hght', self.divide_by_20),
            # annotation => an
            'annotation': ('an', 'annotation', self.default_func),
            # border => bd
            'trbrdrh': ('bd', 'bor-t-r-hi', self.default_func),
            'trbrdrv': ('bd', 'bor-t-r-vi', self.default_func),
            'trbrdrt': ('bd', 'bor-t-r-to', self.default_func),
            'trbrdrl': ('bd', 'bor-t-r-le', self.default_func),
            'trbrdrb': ('bd', 'bor-t-r-bo', self.default_func),
            'trbrdrr': ('bd', 'bor-t-r-ri', self.default_func),
            'clbrdrb': ('bd', 'bor-cel-bo', self.default_func),
            'clbrdrt': ('bd', 'bor-cel-to', self.default_func),
            'clbrdrl': ('bd', 'bor-cel-le', self.default_func),
            'clbrdrr': ('bd', 'bor-cel-ri', self.default_func),
            'brdrb': ('bd', 'bor-par-bo', self.default_func),
            'brdrt': ('bd', 'bor-par-to', self.default_func),
            'brdrl': ('bd', 'bor-par-le', self.default_func),
            'brdrr': ('bd', 'bor-par-ri', self.default_func),
            'box': ('bd', 'bor-par-bx', self.default_func),
            'chbrdr': ('bd', 'bor-par-bo', self.default_func),
            'brdrbtw': ('bd', 'bor-for-ev', self.default_func),
            'brdrbar': ('bd', 'bor-outsid', self.default_func),
            'brdrnone': ('bd', 'bor-none__<false', self.two_part_func),
            # border type => bt
            'brdrs': ('bt', 'bdr-single', self.default_func),
            'brdrth': ('bt', 'bdr-doubtb', self.default_func),
            'brdrsh': ('bt', 'bdr-shadow', self.default_func),
            'brdrdb': ('bt', 'bdr-double', self.default_func),
            'brdrdot': ('bt', 'bdr-dotted', self.default_func),
            'brdrdash': ('bt', 'bdr-dashed', self.default_func),
            'brdrhair': ('bt', 'bdr-hair__', self.default_func),
            'brdrinset': ('bt', 'bdr-inset_', self.default_func),
            'brdrdashsm': ('bt', 'bdr-das-sm', self.default_func),
            'brdrdashd': ('bt', 'bdr-dot-sm', self.default_func),
            'brdrdashdd': ('bt', 'bdr-dot-do', self.default_func),
            'brdroutset': ('bt', 'bdr-outset', self.default_func),
            'brdrtriple': ('bt', 'bdr-trippl', self.default_func),
            'brdrtnthsg': ('bt', 'bdr-thsm__', self.default_func),
            'brdrthtnsg': ('bt', 'bdr-htsm__', self.default_func),
            'brdrtnthtnsg': ('bt', 'bdr-hthsm_', self.default_func),
            'brdrtnthmg': ('bt', 'bdr-thm___', self.default_func),
            'brdrthtnmg': ('bt', 'bdr-htm___', self.default_func),
            'brdrtnthtnmg': ('bt', 'bdr-hthm__', self.default_func),
            'brdrtnthlg': ('bt', 'bdr-thl___', self.default_func),
            'brdrtnthtnlg': ('bt', 'bdr-hthl__', self.default_func),
            'brdrwavy': ('bt', 'bdr-wavy__', self.default_func),
            'brdrwavydb': ('bt', 'bdr-d-wav_', self.default_func),
            'brdrdashdotstr': ('bt', 'bdr-strip_', self.default_func),
            'brdremboss': ('bt', 'bdr-embos_', self.default_func),
            'brdrengrave': ('bt', 'bdr-engra_', self.default_func),
            'brdrframe': ('bt', 'bdr-frame_', self.default_func),
            'brdrw': ('bt', 'bdr-li-wid', self.divide_by_20),
            'brsp': ('bt', 'bdr-sp-wid', self.divide_by_20),
            'brdrcf': ('bt', 'bdr-color_', self.default_func),
            # comments
            # 'comment'            : ('cm', 'comment___', self.default_func),
        }
        self.__number_type_dict = {
            0: 'Arabic',
            1: 'uppercase Roman numeral',
            2: 'lowercase Roman numeral',
            3: 'uppercase letter',
            4: 'lowercase letter',
            5: 'ordinal number',
            6: 'cardianl text number',
            7: 'ordinal text number',
            10: 'Kanji numbering without the digit character',
            11: 'Kanji numbering with the digit character',
            1246: 'phonetic Katakana characters in aiueo order',
            1346: 'phonetic katakana characters in iroha order',
            14: 'double byte character',
            15: 'single byte character',
            16: 'Kanji numbering 3',
            17: 'Kanji numbering 4',
            18: 'Circle numbering',
            19: 'double-byte Arabic numbering',
            2046: 'phonetic double-byte Katakana characters',
            2146: 'phonetic double-byte katakana characters',
            22: 'Arabic with leading zero',
            23: 'bullet',
            24: 'Korean numbering 2',
            25: 'Korean numbering 1',
            26: 'Chinese numbering 1',
            27: 'Chinese numbering 2',
            28: 'Chinese numbering 3',
            29: 'Chinese numbering 4',
            30: 'Chinese Zodiac numbering 1',
            31: 'Chinese Zodiac numbering 2',
            32: 'Chinese Zodiac numbering 3',
            33: 'Taiwanese double-byte numbering 1',
            34: 'Taiwanese double-byte numbering 2',
            35: 'Taiwanese double-byte numbering 3',
            36: 'Taiwanese double-byte numbering 4',
            37: 'Chinese double-byte numbering 1',
            38: 'Chinese double-byte numbering 2',
            39: 'Chinese double-byte numbering 3',
            40: 'Chinese double-byte numbering 4',
            41: 'Korean double-byte numbering 1',
            42: 'Korean double-byte numbering 2',
            43: 'Korean double-byte numbering 3',
            44: 'Korean double-byte numbering 4',
            45: 'Hebrew non-standard decimal',
            46: 'Arabic Alif Ba Tah',
            47: 'Hebrew Biblical standard',
            48: 'Arabic Abjad style',
            255: 'No number',
        }
        self.__language_dict = {
            1078: 'Afrikaans',
            1052: 'Albanian',
            1025: 'Arabic',
            5121: 'Arabic Algeria',
            15361: 'Arabic Bahrain',
            3073: 'Arabic Egypt',
            1: 'Arabic General',
            2049: 'Arabic Iraq',
            11265: 'Arabic Jordan',
            13313: 'Arabic Kuwait',
            12289: 'Arabic Lebanon',
            4097: 'Arabic Libya',
            6145: 'Arabic Morocco',
            8193: 'Arabic Oman',
            16385: 'Arabic Qatar',
            10241: 'Arabic Syria',
            7169: 'Arabic Tunisia',
            14337: 'Arabic U.A.E.',
            9217: 'Arabic Yemen',
            1067: 'Armenian',
            1101: 'Assamese',
            2092: 'Azeri Cyrillic',
            1068: 'Azeri Latin',
            1069: 'Basque',
            1093: 'Bengali',
            4122: 'Bosnia Herzegovina',
            1026: 'Bulgarian',
            1109: 'Burmese',
            1059: 'Byelorussian',
            1027: 'Catalan',
            2052: 'Chinese China',
            4: 'Chinese General',
            3076: 'Chinese Hong Kong',
            4100: 'Chinese Singapore',
            1028: 'Chinese Taiwan',
            1050: 'Croatian',
            1029: 'Czech',
            1030: 'Danish',
            2067: 'Dutch Belgium',
            1043: 'Dutch Standard',
            3081: 'English Australia',
            10249: 'English Belize',
            2057: 'English British',
            4105: 'English Canada',
            9225: 'English Caribbean',
            9: 'English General',
            6153: 'English Ireland',
            8201: 'English Jamaica',
            5129: 'English New Zealand',
            13321: 'English Philippines',
            7177: 'English South Africa',
            11273: 'English Trinidad',
            1033: 'English United States',
            1061: 'Estonian',
            1080: 'Faerose',
            1065: 'Farsi',
            1035: 'Finnish',
            1036: 'French',
            2060: 'French Belgium',
            11276: 'French Cameroon',
            3084: 'French Canada',
            12300: "French Cote d'Ivoire",
            5132: 'French Luxembourg',
            13324: 'French Mali',
            6156: 'French Monaco',
            8204: 'French Reunion',
            10252: 'French Senegal',
            4108: 'French Swiss',
            7180: 'French West Indies',
            9228: 'French Democratic Republic of the Congo',
            1122: 'Frisian',
            1084: 'Gaelic',
            2108: 'Gaelic Ireland',
            1110: 'Galician',
            1079: 'Georgian',
            1031: 'German',
            3079: 'German Austrian',
            5127: 'German Liechtenstein',
            4103: 'German Luxembourg',
            2055: 'German Switzerland',
            1032: 'Greek',
            1095: 'Gujarati',
            1037: 'Hebrew',
            1081: 'Hindi',
            1038: 'Hungarian',
            1039: 'Icelandic',
            1057: 'Indonesian',
            1040: 'Italian',
            2064: 'Italian Switzerland',
            1041: 'Japanese',
            1099: 'Kannada',
            1120: 'Kashmiri',
            2144: 'Kashmiri India',
            1087: 'Kazakh',
            1107: 'Khmer',
            1088: 'Kirghiz',
            1111: 'Konkani',
            1042: 'Korean',
            2066: 'Korean Johab',
            1108: 'Lao',
            1062: 'Latvian',
            1063: 'Lithuanian',
            2087: 'Lithuanian Classic',
            1086: 'Malay',
            2110: 'Malay Brunei Darussalam',
            1100: 'Malayalam',
            1082: 'Maltese',
            1112: 'Manipuri',
            1102: 'Marathi',
            1104: 'Mongolian',
            1121: 'Nepali',
            2145: 'Nepali India',
            1044: 'Norwegian Bokmal',
            2068: 'Norwegian Nynorsk',
            1096: 'Oriya',
            1045: 'Polish',
            1046: 'Portuguese (Brazil)',
            2070: 'Portuguese (Portugal)',
            1094: 'Punjabi',
            1047: 'Rhaeto-Romanic',
            1048: 'Romanian',
            2072: 'Romanian Moldova',
            1049: 'Russian',
            2073: 'Russian Moldova',
            1083: 'Sami Lappish',
            1103: 'Sanskrit',
            3098: 'Serbian Cyrillic',
            2074: 'Serbian Latin',
            1113: 'Sindhi',
            1051: 'Slovak',
            1060: 'Slovenian',
            1070: 'Sorbian',
            11274: 'Spanish Argentina',
            16394: 'Spanish Bolivia',
            13322: 'Spanish Chile',
            9226: 'Spanish Colombia',
            5130: 'Spanish Costa Rica',
            7178: 'Spanish Dominican Republic',
            12298: 'Spanish Ecuador',
            17418: 'Spanish El Salvador',
            4106: 'Spanish Guatemala',
            18442: 'Spanish Honduras',
            2058: 'Spanish Mexico',
            3082: 'Spanish Modern',
            19466: 'Spanish Nicaragua',
            6154: 'Spanish Panama',
            15370: 'Spanish Paraguay',
            10250: 'Spanish Peru',
            20490: 'Spanish Puerto Rico',
            1034: 'Spanish Traditional',
            14346: 'Spanish Uruguay',
            8202: 'Spanish Venezuela',
            1072: 'Sutu',
            1089: 'Swahili',
            1053: 'Swedish',
            2077: 'Swedish Finland',
            1064: 'Tajik',
            1097: 'Tamil',
            1092: 'Tatar',
            1098: 'Telugu',
            1054: 'Thai',
            1105: 'Tibetan',
            1073: 'Tsonga',
            1074: 'Tswana',
            1055: 'Turkish',
            1090: 'Turkmen',
            1058: 'Ukranian',
            1056: 'Urdu',
            2080: 'Urdu India',
            2115: 'Uzbek Cyrillic',
            1091: 'Uzbek Latin',
            1075: 'Venda',
            1066: 'Vietnamese',
            1106: 'Welsh',
            1076: 'Xhosa',
            1085: 'Yiddish',
            1077: 'Zulu',
            1024: 'Unkown',
            255: 'Unkown',
        }
        # unknown
        # These must get passed on because they occurred after \\*
        # 'do'                :   ('un', 'unknown___', self.default_func),
        # 'company'           :   ('un', 'company___', self.default_func),
        # 'shpinst'           :   ('un', 'unknown___', self.default_func),
        # 'panose'            :   ('un', 'unknown___', self.default_func),
        # 'falt'              :   ('un', 'unknown___', self.default_func),
        # 'listoverridetable' :   ('un', 'unknown___', self.default_func),
        # 'category'          :   ('un', 'unknown___', self.default_func),
        # 'template'          :   ('un', 'unknown___', self.default_func),
        # 'ud'                :   ('un', 'unknown___', self.default_func),
        # 'formfield'         :   ('un', 'unknown___', self.default_func),
        # 'ts'                :   ('un', 'unknown___', self.default_func),
        # 'rsidtbl'           :   ('un', 'unknown___', self.default_func),
        # 'generator'         :   ('un', 'unknown___', self.default_func),
        # 'ftnsep'            :   ('un', 'unknown___', self.default_func),
        # 'aftnsep'           :   ('un', 'unknown___', self.default_func),
        # 'aftnsepc'          :   ('un', 'unknown___', self.default_func),
        # 'aftncn'            :   ('un', 'unknown___', self.default_func),
        # 'objclass'          :   ('un', 'unknown___', self.default_func),
        # 'objdata'           :   ('un', 'unknown___', self.default_func),
        # 'picprop'           :   ('un', 'unknown___', self.default_func),
        # 'blipuid'           :   ('un', 'unknown___', self.default_func),

    def __ms_hex_func(self, pre, token, num):
        num = num[1:]  # chop off leading 0, which I added
        num = num.upper()  # the mappings store hex in caps
        return f"tx<hx<__________<'{num}\n"  # add an ' for the mappings

    def ms_sub_func(self, pre, token, num):
        return f'tx<mc<__________<{token}\n'

    def direct_conv_func(self, pre, token, num):
        return f'mi<tg<empty_____<{token}\n'

    def default_func(self, pre, token, num):
        if num is None:
            num = 'true'
        return f'cw<{pre}<{token}<nu<{num}\n'

    def colorz_func(self, pre, token, num):
        if num is None:
            num = '0'
        return f'cw<{pre}<{token}<nu<{num}\n'

    def __list_type_func(self, pre, token, num):
        type = 'arabic'
        if num is None:
            type = 'Arabic'
        else:
            try:
                num = int(num)
            except ValueError:
                if self.__run_level > 3:
                    msg = f'Number "{num}" cannot be converted to integer\n'
                    raise self.__bug_handler(msg)
            type = self.__number_type_dict.get(num)
            if type is None:
                if self.__run_level > 3:
                    msg = 'No type for "%s" in self.__number_type_dict\n'
                    raise self.__bug_handler
                type = 'Arabic'
        return f'cw<{pre}<{token}<nu<{type}\n'

    def __language_func(self, pre, token, num):
        _lang_m = re.search(r'[0-9]+', num)
        assert _lang_m is not None
        lang_name = self.__language_dict.get(int(_lang_m.group()))
        if not lang_name:
            lang_name = 'not defined'
            if self.__run_level > 3:
                msg = f'No entry for number "{num}"'
                raise self.__bug_handler(msg)
        return f'cw<{pre}<{token}<nu<{lang_name}\n'

    def two_part_func(self, pre, token, num):
        list = token.split('<')
        token = list[0]
        num = list[1]
        return f'cw<{pre}<{token}<nu<{num}\n'
        # return 'cw<nu<nu<nu<%s>num<%s\n' % (token, num)

    def divide_by_2(self, pre, token, num):
        num = self.divide_num(num, 2)
        return f'cw<{pre}<{token}<nu<{num}\n'
        # return 'cw<nu<nu<nu<%s>%s<%s\n' % (token, num, token)

    def divide_by_20(self, pre, token, num):
        num = self.divide_num(num, 20)
        return f'cw<{pre}<{token}<nu<{num}\n'
        # return 'cw<nu<nu<nu<%s>%s<%s\n' % (token, num, token)

    def text_func(self, pre, token, num=None):
        return f'tx<nu<__________<{token}\n'

    def ob_func(self, pre, token, num=None):
        self.__bracket_count += 1
        return f'ob<nu<open-brack<{self.__bracket_count:04}\n'

    def cb_func(self, pre, token, num=None):
        line = f'cb<nu<clos-brack<{self.__bracket_count:04}\n'
        self.__bracket_count -= 1
        return line

    def color_func(self, pre, token, num):
        third_field = 'nu'
        if num[-1] == ';':
            num = num[:-1]
            third_field = 'en'
        num = f'{int(num):X}'
        if len(num) != 2:
            num = '0' + num
        return f'cw<{pre}<{token}<{third_field}<{num}\n'
        # return 'cw<cl<%s<nu<nu<%s>%s<%s\n' % (third_field, token, num, token)

    def bool_st_func(self, pre, token, num):
        if num is None or num in {'', '1'}:
            return f'cw<{pre}<{token}<nu<true\n'
            # return 'cw<nu<nu<nu<%s>true<%s\n' % (token, token)
        elif num == '0':
            return f'cw<{pre}<{token}<nu<false\n'
            # return 'cw<nu<nu<nu<%s>false<%s\n' % (token, token)
        else:
            msg = f"boolean should have some value module process tokens\ntoken is {token}\n'{num}'\n"
            raise self.__bug_handler(msg)

    def __no_sup_sub_func(self, pre, token, num):
        the_string = 'cw<ci<subscript_<nu<false\n'
        the_string += 'cw<ci<superscrip<nu<false\n'
        return the_string

    def divide_num(self, numerator, denominator):
        try:
            # calibre why ignore negative number? Wrong in case of \fi
            _num_m = re.search(r'[0-9.\-]+', numerator)
            assert _num_m is not None
            numerator = float(_num_m.group())
        except TypeError:
            if self.__run_level > 3:
                msg = (
                    'No number to process?\nthis indicates that the token \\(\\li\\)'
                    'should have a number and does not\nnumerator is'
                    f'"{numerator}"\ndenominator is "{denominator}"\n'
                )
                raise self.__bug_handler(msg)
            self.__return_code = max(self.__return_code, 5)
            return 0
        num = f'{round(numerator / denominator, 2):0.2f}'
        return num
        string_num = str(num)
        if string_num[-2:] == '.0':
            string_num = string_num[:-2]
        return string_num

    def split_let_num(self, token):
        match_obj = re.search(self.__num_exp, token)
        if match_obj is not None:
            first = match_obj.group(1)
            second = match_obj.group(2)
            if not second:
                if self.__run_level > 3:
                    msg = f"token is '{token}' \n"
                    raise self.__bug_handler(msg)
                return first, 0
        else:
            if self.__run_level > 3:
                msg = f"token is '{token}' \n"
                raise self.__bug_handler
            return token, 0
        return first, second

    def convert_to_hex(self, number):
        """Convert a string to uppercase hexadecimal"""
        num = int(number)
        try:
            hex_num = f'{num:X}'
            return hex_num
        except Exception:
            raise self.__bug_handler

    def process_cw(self, token):
        """Change the value of the control word by determining what dictionary
        it belongs to"""
        special = ['*', ':', '}', '{', '~', '_', '-', ';']
        # if token != "{" or token != "}":
        token = token[1:]  # strip off leading \
        token = token.replace(' ', '')
        # if not token: return
        only_alpha = token.isalpha()
        num = None
        if not only_alpha and token not in special:
            token, num = self.split_let_num(token)
        pre, token, action = self.dict_token.get(token, (None, None, None))
        if action:
            return action(pre, token, num)

    def __check_brackets(self, in_file):
        self.__check_brack_obj = check_brackets.CheckBrackets(file=in_file)
        good_br = self.__check_brack_obj.check_brackets()[0]
        if not good_br:
            return 1

    def process_tokens(self):
        """Main method for handling other methods."""
        line_count = 0
        with open_for_read(self.__file) as read_obj:
            with open_for_write(self.__write_to) as write_obj:
                for line in read_obj:
                    token = line.replace('\n', '')
                    line_count += 1
                    if line_count == 1 and token != '\\{':
                        msg = "\nInvalid RTF: document doesn't start with {\n"
                        raise self.__exception_handler(msg)
                    elif line_count == 2 and token[0:4] != '\\rtf':
                        msg = "\nInvalid RTF: document doesn't start with \\rtf \n"
                        raise self.__exception_handler(msg)

                    the_index = token.find('\\ ')
                    if token is not None and the_index > -1:
                        msg = f'\nInvalid RTF: token "\\ " not valid.\nError at line {line_count}'
                        raise self.__exception_handler(msg)
                    elif token[:1] == '\\':
                        line = self.process_cw(token)
                        if line is not None:
                            write_obj.write(line)
                    else:
                        fields = re.split(self.__utf_exp, token)
                        for field in fields:
                            if not field:
                                continue
                            if field[0:1] == '&':
                                write_obj.write(f'tx<ut<__________<{field}\n')
                            else:
                                write_obj.write(f'tx<nu<__________<{field}\n')

        if not line_count:
            msg = '\nInvalid RTF: file appears to be empty.\n'
            raise self.__exception_handler(msg)

        copy_obj = copy.Copy(bug_handler=self.__bug_handler)
        if self.__copy:
            copy_obj.copy_file(self.__write_to, 'processed_tokens.data')
        copy_obj.rename(self.__write_to, self.__file)
        os.remove(self.__write_to)

        bad_brackets = self.__check_brackets(self.__file)
        if bad_brackets:
            msg = '\nInvalid RTF: document does not have matching brackets.\n'
            raise self.__exception_handler(msg)
        else:
            return self.__return_code



# License: GPLv3 Copyright: 2011, Leigh Parry <leighparry@blueyonder.co.uk>

import re


def unsmarten(txt):
    txt = re.sub(r'&#162;|&cent;|¢', r'{c\}', txt)  # cent
    txt = re.sub(r'&#163;|&pound;|£', r'{L-}', txt)  # pound
    txt = re.sub(r'&#165;|&yen;|¥', r'{Y=}', txt)  # yen
    txt = re.sub(r'&#169;|&copy;|©', r'{(c)}', txt)  # copyright
    txt = re.sub(r'&#174;|&reg;|®', r'{(r)}', txt)  # registered
    txt = re.sub(r'&#188;|&frac14;|¼', r'{1/4}', txt)  # quarter
    txt = re.sub(r'&#189;|&frac12;|½', r'{1/2}', txt)  # half
    txt = re.sub(r'&#190;|&frac34;|¾', r'{3/4}', txt)  # three-quarter
    txt = re.sub(r'&#192;|&Agrave;|À', r'{A`)}', txt)  # A-grave
    txt = re.sub(r'&#193;|&Aacute;|Á', r"{A'}", txt)  # A-acute
    txt = re.sub(r'&#194;|&Acirc;|Â', r'{A^}', txt)  # A-circumflex
    txt = re.sub(r'&#195;|&Atilde;|Ã', r'{A~}', txt)  # A-tilde
    txt = re.sub(r'&#196;|&Auml;|Ä', r'{A"}', txt)  # A-umlaut
    txt = re.sub(r'&#197;|&Aring;|Å', r'{Ao}', txt)  # A-ring
    txt = re.sub(r'&#198;|&AElig;|Æ', r'{AE}', txt)  # AE
    txt = re.sub(r'&#199;|&Ccedil;|Ç', r'{C,}', txt)  # C-cedilla
    txt = re.sub(r'&#200;|&Egrave;|È', r'{E`}', txt)  # E-grave
    txt = re.sub(r'&#201;|&Eacute;|É', r"{E'}", txt)  # E-acute
    txt = re.sub(r'&#202;|&Ecirc;|Ê', r'{E^}', txt)  # E-circumflex
    txt = re.sub(r'&#203;|&Euml;|Ë', r'{E"}', txt)  # E-umlaut
    txt = re.sub(r'&#204;|&Igrave;|Ì', r'{I`}', txt)  # I-grave
    txt = re.sub(r'&#205;|&Iacute;|Í', r"{I'}", txt)  # I-acute
    txt = re.sub(r'&#206;|&Icirc;|Î', r'{I^}', txt)  # I-circumflex
    txt = re.sub(r'&#207;|&Iuml;|Ï', r'{I"}', txt)  # I-umlaut
    txt = re.sub(r'&#208;|&ETH;|Ð', r'{D-}', txt)  # ETH
    txt = re.sub(r'&#209;|&Ntilde;|Ñ', r'{N~}', txt)  # N-tilde
    txt = re.sub(r'&#210;|&Ograve;|Ò', r'{O`}', txt)  # O-grave
    txt = re.sub(r'&#211;|&Oacute;|Ó', r"{O'}", txt)  # O-acute
    txt = re.sub(r'&#212;|&Ocirc;|Ô', r'{O^}', txt)  # O-circumflex
    txt = re.sub(r'&#213;|&Otilde;|Õ', r'{O~}', txt)  # O-tilde
    txt = re.sub(r'&#214;|&Ouml;|Ö', r'{O"}', txt)  # O-umlaut
    txt = re.sub(r'&#215;|&times;|×', r'{x}', txt)  # dimension
    txt = re.sub(r'&#216;|&Oslash;|Ø', r'{O/}', txt)  # O-slash
    txt = re.sub(r'&#217;|&Ugrave;|Ù', r'{U`}', txt)  # U-grave
    txt = re.sub(r'&#218;|&Uacute;|Ú', r"{U'}", txt)  # U-acute
    txt = re.sub(r'&#219;|&Ucirc;|Û', r'{U^}', txt)  # U-circumflex
    txt = re.sub(r'&#220;|&Uuml;|Ü', r'{U"}', txt)  # U-umlaut
    txt = re.sub(r'&#221;|&Yacute;|Ý', r"{Y'}", txt)  # Y-grave
    txt = re.sub(r'&#223;|&szlig;|ß', r'{sz}', txt)  # sharp-s
    txt = re.sub(r'&#224;|&agrave;|à', r'{a`}', txt)  # a-grave
    txt = re.sub(r'&#225;|&aacute;|á', r"{a'}", txt)  # a-acute
    txt = re.sub(r'&#226;|&acirc;|â', r'{a^}', txt)  # a-circumflex
    txt = re.sub(r'&#227;|&atilde;|ã', r'{a~}', txt)  # a-tilde
    txt = re.sub(r'&#228;|&auml;|ä', r'{a"}', txt)  # a-umlaut
    txt = re.sub(r'&#229;|&aring;|å', r'{ao}', txt)  # a-ring
    txt = re.sub(r'&#230;|&aelig;|æ', r'{ae}', txt)  # ae
    txt = re.sub(r'&#231;|&ccedil;|ç', r'{c,}', txt)  # c-cedilla
    txt = re.sub(r'&#232;|&egrave;|è', r'{e`}', txt)  # e-grave
    txt = re.sub(r'&#233;|&eacute;|é', r"{e'}", txt)  # e-acute
    txt = re.sub(r'&#234;|&ecirc;|ê', r'{e^}', txt)  # e-circumflex
    txt = re.sub(r'&#235;|&euml;|ë', r'{e"}', txt)  # e-umlaut
    txt = re.sub(r'&#236;|&igrave;|ì', r'{i`}', txt)  # i-grave
    txt = re.sub(r'&#237;|&iacute;|í', r"{i'}", txt)  # i-acute
    txt = re.sub(r'&#238;|&icirc;|î', r'{i^}', txt)  # i-circumflex
    txt = re.sub(r'&#239;|&iuml;|ï', r'{i"}', txt)  # i-umlaut
    txt = re.sub(r'&#240;|&eth;|ð', r'{d-}', txt)  # eth
    txt = re.sub(r'&#241;|&ntilde;|ñ', r'{n~}', txt)  # n-tilde
    txt = re.sub(r'&#242;|&ograve;|ò', r'{o`}', txt)  # o-grave
    txt = re.sub(r'&#243;|&oacute;|ó', r"{o'}", txt)  # o-acute
    txt = re.sub(r'&#244;|&ocirc;|ô', r'{o^}', txt)  # o-circumflex
    txt = re.sub(r'&#245;|&otilde;|õ', r'{o~}', txt)  # o-tilde
    txt = re.sub(r'&#246;|&ouml;|ö', r'{o"}', txt)  # o-umlaut
    txt = re.sub(r'&#248;|&oslash;|ø', r'{o/}', txt)  # o-stroke
    txt = re.sub(r'&#249;|&ugrave;|ù', r'{u`}', txt)  # u-grave
    txt = re.sub(r'&#250;|&uacute;|ú', r"{u'}", txt)  # u-acute
    txt = re.sub(r'&#251;|&ucirc;|û', r'{u^}', txt)  # u-circumflex
    txt = re.sub(r'&#252;|&uuml;|ü', r'{u"}', txt)  # u-umlaut
    txt = re.sub(r'&#253;|&yacute;|ý', r"{y'}", txt)  # y-acute
    txt = re.sub(r'&#255;|&yuml;|ÿ', r'{y"}', txt)  # y-umlaut

    txt = re.sub(r'&#268;|&Ccaron;|Č', r'{Cˇ}', txt)  # C-caron
    txt = re.sub(r'&#269;|&ccaron;|č', r'{cˇ}', txt)  # c-caron
    txt = re.sub(r'&#270;|&Dcaron;|Ď', r'{Dˇ}', txt)  # D-caron
    txt = re.sub(r'&#271;|&dcaron;|ď', r'{dˇ}', txt)  # d-caron
    txt = re.sub(r'&#282;|&Ecaron;|Ě', r'{Eˇ}', txt)  # E-caron
    txt = re.sub(r'&#283;|&ecaron;|ě', r'{eˇ}', txt)  # e-caron
    txt = re.sub(r'&#313;|&Lacute;|Ĺ', r"{L'}", txt)  # L-acute
    txt = re.sub(r'&#314;|&lacute;|ĺ', r"{l'}", txt)  # l-acute
    txt = re.sub(r'&#317;|&Lcaron;|Ľ', r'{Lˇ}', txt)  # L-caron
    txt = re.sub(r'&#318;|&lcaron;|ľ', r'{lˇ}', txt)  # l-caron
    txt = re.sub(r'&#327;|&Ncaron;|Ň', r'{Nˇ}', txt)  # N-caron
    txt = re.sub(r'&#328;|&ncaron;|ň', r'{nˇ}', txt)  # n-caron

    txt = re.sub(r'&#338;|&OElig;|Œ', r'{OE}', txt)  # OE
    txt = re.sub(r'&#339;|&oelig;|œ', r'{oe}', txt)  # oe

    txt = re.sub(r'&#340;|&Racute;|Ŕ', r"{R'}", txt)  # R-acute
    txt = re.sub(r'&#341;|&racute;|ŕ', r"{r'}", txt)  # r-acute
    txt = re.sub(r'&#344;|&Rcaron;|Ř', r'{Rˇ}', txt)  # R-caron
    txt = re.sub(r'&#345;|&rcaron;|ř', r'{rˇ}', txt)  # r-caron
    txt = re.sub(r'&#348;|Ŝ', r'{S^}', txt)  # S-circumflex
    txt = re.sub(r'&#349;|ŝ', r'{s^}', txt)  # s-circumflex
    txt = re.sub(r'&#352;|&Scaron;|Š', r'{Sˇ}', txt)  # S-caron
    txt = re.sub(r'&#353;|&scaron;|š', r'{sˇ}', txt)  # s-caron
    txt = re.sub(r'&#356;|&Tcaron;|Ť', r'{Tˇ}', txt)  # T-caron
    txt = re.sub(r'&#357;|&tcaron;|ť', r'{tˇ}', txt)  # t-caron
    txt = re.sub(r'&#366;|&Uring;|Ů', r'{U°}', txt)  # U-ring
    txt = re.sub(r'&#367;|&uring;|ů', r'{u°}', txt)  # u-ring
    txt = re.sub(r'&#381;|&Zcaron;|Ž', r'{Zˇ}', txt)  # Z-caron
    txt = re.sub(r'&#382;|&zcaron;|ž', r'{zˇ}', txt)  # z-caron

    txt = re.sub(r'&#8226;|&bull;|•', r'{*}', txt)  # bullet
    txt = re.sub(r'&#8355;|₣', r'{Fr}', txt)  # Franc
    txt = re.sub(r'&#8356;|₤', r'{L=}', txt)  # Lira
    txt = re.sub(r'&#8360;|₨', r'{Rs}', txt)  # Rupee
    txt = re.sub(r'&#8364;|&euro;|€', r'{C=}', txt)  # euro
    txt = re.sub(r'&#8482;|&trade;|™', r'{tm}', txt)  # trademark
    txt = re.sub(r'&#9824;|&spades;|♠', r'{spade}', txt)  # spade
    txt = re.sub(r'&#9827;|&clubs;|♣', r'{club}', txt)  # club
    txt = re.sub(r'&#9829;|&hearts;|♥', r'{heart}', txt)  # heart
    txt = re.sub(r'&#9830;|&diams;|♦', r'{diamond}', txt)  # diamond

    # Move into main code?
    # txt = re.sub(r'\xa0',     r'p. ',         txt)  # blank paragraph
    # txt = re.sub(r'\n\n\n\n', r'\n\np. \n\n', txt)  # blank paragraph
    # txt = re.sub(r'\n  \n',   r'\n<br />\n',  txt)  # blank paragraph - br tag

    return txt
